(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Check_Conditional_CRP
imports 
  Check_CRP
begin

subsection \<open>Proving that a conditional TRS is quasi reductive\<close>

datatype ('f, 'l, 'v) conditional_cr_proof =
  Unravel_CR 
    "((('f,'l)lab,'v)crule \<times> (('f,'l)lab,'v) rules) list"
    "('f, 'l, 'v) cr_proof"

primrec
  check_conditional_cr_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,key,countable}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> (('f,label_type)lab,string)crules \<Rightarrow> 
     ('f, label_type, string) conditional_cr_proof \<Rightarrow>
     shows check" 
where
  "check_conditional_cr_proof a i I J ctrs (Unravel_CR u_info prf) = debug (i []) ''Unravel'' (do {
    r \<leftarrow> check_sp_unraveling u_info ctrs
      <+? (\<lambda> s. i +@+ '': error in unraveling'' +#+ shows_nl +@+ indent s);
    check_cr_proof a (i +@+ shows_string ''.1'') I J r prf
      <+? (\<lambda>s. i +@+ '': error below confluence proof'' +#+ shows_nl +@+ indent s)
  })"

primrec conditional_cr_assms :: "bool \<Rightarrow> ('f, 'l, 'v) conditional_cr_proof \<Rightarrow> (('f, 'l, 'v) assm) list" where
  "conditional_cr_assms a (Unravel_CR _ p) = cr_assms a p"

lemma check_conditional_cr_proof_with_assms_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (conditional_cr_assms a prf). holds p"
    and ok: "isOK (check_conditional_cr_proof a i I J ctrs prf)"
  shows "CR (cstep (set ctrs))"
  using fin ok
proof -
  interpret tp_spec I by fact
  from ok fin show ?thesis
  proof (induct "prf" arbitrary: i)
    case (Unravel_CR u_info prof)
    note check = Unravel_CR(1)[simplified]
    from check obtain R where R: "check_sp_unraveling u_info ctrs = return R" by auto
    from check[unfolded R]
    have ok: "isOK (check_cr_proof a (i +@+ shows_string ''.1'') I J R prof)" by simp
    from Unravel_CR(2) have a: "\<forall>a\<in>set (cr_assms a prof). holds a" by auto
    from check_cr_proof_with_assms_sound[OF I J a ok] have CR: "CR (rstep (set R))" by simp
    show ?case
      by (rule check_sp_unraveling_CR[OF R infinite_lab CR])
  qed
qed

lemma conditional_cr_assms_False[simp]: "conditional_cr_assms False prf = []"
  by (induct "prf") simp_all

lemma check_conditional_cr_proof_sound: 
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and ok: "isOK (check_conditional_cr_proof False i I J ctrs prf)"
  shows "CR (cstep (set ctrs))"
  by (rule check_conditional_cr_proof_with_assms_sound[OF I J _ ok], simp)

end
