(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Conditional_Rewriting
imports
  "../Rewriting/Trs"
  "../Orderings/Reduction_Pair"
begin

type_synonym ('f, 'v) condition = "('f, 'v) term \<times> ('f, 'v) term"
type_synonym ('f, 'v) crule = "('f, 'v) rule \<times> ('f, 'v) condition list"
type_synonym ('f, 'v) ctrs = "('f, 'v) crule set"

text \<open>A conditional rewrite step of level @{term n}.\<close>
fun cstep_n :: "('f, 'v) ctrs \<Rightarrow> nat \<Rightarrow> ('f, 'v) term rel"
where
  "cstep_n R 0 = {}" |
  cstep_n_Suc: "cstep_n R (Suc n) =
    {(C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) | C l r \<sigma> cs.
      ((l, r), cs) \<in> R \<and> (\<forall> (s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*)}"

definition conds_sat
where
  conds_sat_iff: "conds_sat R n cs \<sigma> \<longleftrightarrow> (\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*)"

lemma conds_sat_append [simp]:
  "conds_sat R n (cs @ ds) \<sigma> \<longleftrightarrow> conds_sat R n cs \<sigma> \<and> conds_sat R n ds \<sigma>"
by (auto simp add: conds_sat_iff)

lemma cons_sat_singleton [simp]:
  "conds_sat R n [c] \<sigma> \<longleftrightarrow> (fst c \<cdot> \<sigma>, snd c \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*"
by (auto simp: conds_sat_iff)

lemma cstep_n_SucI [Pure.intro?]:
  assumes "((l, r), cs) \<in> R"
    and "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*"
    and "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  shows "(s, t) \<in> cstep_n R (Suc n)"
  using assms by (auto) blast

lemma cstep_nE:
  assumes "(s, t) \<in> cstep_n R n"
  obtains C l r \<sigma> cs n' where "((l, r), cs) \<in> R" and "n = Suc n'"
    and "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n')\<^sup>*"
    and "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  using assms by (cases n) auto

lemma cstep_n_SucE:
  assumes "(s, t) \<in> cstep_n R (Suc n)"
  obtains C l r \<sigma> cs where "((l, r), cs) \<in> R"
    and "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*"
    and "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  using assms by auto

declare cstep_n_Suc [simp del]

lemma cstep_n_Suc_mono:
  "cstep_n R n \<subseteq> cstep_n R (Suc n)"
proof (induct n)
  case (Suc n)
  show ?case
    using rtrancl_mono [OF Suc] by (auto elim!: cstep_n_SucE intro!: cstep_n_SucI)
qed simp

lemma cstep_n_mono:
  assumes "i \<le> n"
  shows "cstep_n R i \<subseteq> cstep_n R n"
using assms
proof (induct "n - i" arbitrary: i)
  case (Suc k)
  then have "k = n - Suc i" and "Suc i \<le> n" by arith+
  from Suc.hyps(1) [OF this] show ?case using cstep_n_Suc_mono by blast
qed simp

lemma cstep_n_mono':
  "cstep_n R m \<subseteq> cstep_n R (m + k)"
  using cstep_n_mono [of "m" "m + k"] by simp

lemma cstep_n_ctxt:
  assumes "(s, t) \<in> cstep_n R n"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> cstep_n R n"
  using assms by (cases n) (auto elim!: cstep_n_SucE intro: cstep_n_SucI [where C = "C \<circ>\<^sub>c D" for D])

lemma cstep_n_subst:
  assumes "(s, t) \<in> cstep_n R n"
  shows "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> cstep_n R n"
using assms
proof (induct n arbitrary: s t \<sigma>)
  case (Suc n)
  from \<open>(s, t) \<in> cstep_n R (Suc n)\<close> obtain C l r cs \<tau>
    where "((l, r), cs) \<in> R" and "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<tau>, t\<^sub>i \<cdot> \<tau>) \<in> (cstep_n R n)\<^sup>*"
    and "s = C\<langle>l \<cdot> \<tau>\<rangle>" and "t = C\<langle>r \<cdot> \<tau>\<rangle>" by (rule cstep_n_SucE)
  then show ?case
    using Suc.hyps and rtrancl_map [of "cstep_n R n" "\<lambda>t. t \<cdot> \<sigma>"]
    by (auto intro!: cstep_n_SucI [of l r cs R "\<tau> \<circ>\<^sub>s \<sigma>" _ _ "C \<cdot>\<^sub>c \<sigma>"])
qed simp

lemma cstep_n_ctxt_subst:
  assumes "(s, t) \<in> cstep_n R n"
  shows "(D \<langle>s \<cdot> \<sigma>\<rangle>, D \<langle>t \<cdot> \<sigma>\<rangle>) \<in> cstep_n R n"
  by (intro cstep_n_ctxt cstep_n_subst) fact

definition "cstep R = (\<Union>n. cstep_n R n)"

lemma cstep_iff:
  "(s, t) \<in> cstep R \<longleftrightarrow> (\<exists>n. (s, t) \<in> cstep_n R n)"
  by (auto simp: cstep_def)
  
lemma NF_cstep_subterm:
  assumes "t \<in> NF (cstep R)" and "t \<unrhd> s"
  shows "s \<in> NF (cstep R)"
proof (rule ccontr)  
  assume "\<not> ?thesis"
  then obtain u where "(s, u) \<in> cstep R" by auto
  from \<open>t \<unrhd> s\<close> obtain C where "t = C\<langle>s\<rangle>" by auto
  with \<open>(s, u) \<in> cstep R\<close> have "(t, C\<langle>u\<rangle>) \<in> cstep R"
    using cstep_iff cstep_n_ctxt by blast
  hence "t \<notin> NF (cstep R)" by auto
  with assms show False by simp
qed

lemma cstep_n_empty [simp]: "cstep_n {} n = {}"
  by (induct n) (auto simp: cstep_n_Suc)

lemma cstep_empty [simp]: "cstep {} = {}"
  unfolding cstep_def by simp

lemma csteps_n_trans:
  assumes "(s, t) \<in> (cstep_n R m)\<^sup>*" and "(t, u) \<in> (cstep_n R n)\<^sup>*"
  shows "(s, u) \<in> (cstep_n R (max m n))\<^sup>*"
  using assms
    and cstep_n_mono [of m "max m n" R, THEN rtrancl_mono]
    and cstep_n_mono [of n "max m n" R, THEN rtrancl_mono]
    by (auto dest!: set_mp)

lemma csteps_imp_csteps_n:
  assumes "(s, t) \<in> (cstep R)\<^sup>*"
  shows "\<exists>n. (s, t) \<in> (cstep_n R n)\<^sup>*"
  using assms by (induct) (auto intro: csteps_n_trans simp: cstep_iff)

lemma all_cstep_imp_cstep_n:
  assumes "\<forall>i < (k::nat). (s\<^sub>i i, t\<^sub>i i) \<in> (cstep R)\<^sup>*"
  shows "\<exists>n. \<forall>i < k. (s\<^sub>i i, t\<^sub>i i) \<in> (cstep_n R n)\<^sup>*" (is "\<exists>n. \<forall>i < k. ?P i n")
proof -
  have "\<forall>i < k. \<exists>n\<^sub>i. ?P i n\<^sub>i"
    using assms by (auto intro: csteps_imp_csteps_n)
  then obtain f where *: "\<forall>i < k. ?P i (f i)" by metis
  def n \<equiv> "Max (set (map f [0 ..< k]))"
  have "\<forall>i < k. f i \<le> n" by (auto simp: n_def)
  then have "\<forall>i < k. ?P i n"
    using cstep_n_mono [of "f i" n R for i, THEN rtrancl_mono] and * by blast
  then show ?thesis ..
qed

lemma cstepI:
  assumes "((l, r), cs) \<in> R"
    and conds: "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
    and "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
    and "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
  shows "(s, t) \<in> cstep R"
proof -
  obtain n where "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*"
  using all_cstep_imp_cstep_n[of "length cs" "\<lambda>i. fst (cs ! i) \<cdot> \<sigma>" "\<lambda>i. snd (cs ! i) \<cdot> \<sigma>" R]
  and conds
    by (auto simp: all_set_conv_all_nth split_beta')
  then have "(s, t) \<in> cstep_n R (Suc n)" using assms by (intro cstep_n_SucI) (assumption)+
  then show ?thesis using cstep_iff by auto
qed

lemma cstepE [elim]:
  assumes "(s, t) \<in> cstep R"
    and imp: "\<And>C \<sigma> l r cs. \<lbrakk>((l, r), cs) \<in> R; \<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*; s = C\<langle>l \<cdot> \<sigma>\<rangle>; t = C\<langle>r \<cdot> \<sigma>\<rangle>\<rbrakk> \<Longrightarrow> P"
  shows "P"
proof -
  obtain n where *: "(s, t) \<in> cstep_n R n" using assms by (auto simp: cstep_iff)
  then show ?thesis
  proof (cases n)
    case (Suc n')
    have **: "(cstep_n R n')\<^sup>* \<subseteq> (cstep R)\<^sup>*" by (rule rtrancl_mono) (auto simp: cstep_iff)
    show ?thesis
      by (rule cstep_n_SucE [OF * [unfolded Suc]], rule imp, insert **) auto
  qed simp
qed

lemma Var_NF_cstep:
  assumes "\<forall> ((l, r), cs) \<in> R. is_Fun l"
  shows "Var x \<in> NF (cstep R)"
proof
  fix s
  show "(Var x, s) \<notin> cstep R"
  proof
    assume "(Var x, s) \<in> cstep R"
    then show "False" using assms supteq_var_imp_eq by fastforce
  qed
qed

lemma cstep_ctxt_closed:
  "ctxt_closed (cstep R)"
  by (rule ctxt_closedI) (auto intro: cstep_n_ctxt simp: cstep_iff)

lemma csteps_all_ctxt_closed[intro]: "all_ctxt_closed UNIV ((cstep r)^*)"
  using  cstep_ctxt_closed by (blast intro: trans_ctxt_imp_all_ctxt_closed)

lemma join_csteps_all_ctxt_closed: "all_ctxt_closed UNIV ((cstep r)\<^sup>\<down>)" 
proof -
 have var:"\<forall>x. (Var x, Var x) \<in> (cstep r)\<^sup>\<down>" by fast
 { fix f ss ts
   assume a:"length ts = length ss" "\<forall>i<length ts. (ts ! i, ss ! i) \<in> (cstep r)\<^sup>\<down>"
   from a(2) have "\<forall>i<length ts. \<exists> ui. (ts ! i, ui)\<in> (cstep r)^* \<and>  (ss ! i, ui) \<in> (cstep r)^*" by blast
   with choice have "\<exists>ui. \<forall>i<length ts. (ts ! i, ui i)\<in> (cstep r)^* \<and>  (ss ! i, ui i) \<in> (cstep r)^*" by meson
   then obtain ui where ui:"\<And>i. i<length ts \<Longrightarrow> (ts ! i, ui i)\<in> (cstep r)^* \<and>  (ss ! i, ui i) \<in> (cstep r)^*" by blast
   let ?us = "map ui [0 ..<length ts]"
   have lus: "length ts = length ?us" by simp
   from ui have uss:"\<And>j. j < length ts \<Longrightarrow> (ss ! j, ?us ! j) \<in> (cstep r)^*" by auto
   from ui have ust:"\<And>j. j < length ts \<Longrightarrow> (ts ! j, ?us ! j) \<in> (cstep r)^*" by auto
   note acc = conjunct1[OF csteps_all_ctxt_closed[unfolded all_ctxt_closed_def], rule_format]
   from acc[OF _ lus ust] have tseq:"(Fun f ts, Fun f ?us) \<in> (cstep r)\<^sup>*" by blast
   from acc[OF _ lus[unfolded a(1)] uss[unfolded a(1)]] a(1) have sseq:"(Fun f ss, Fun f ?us) \<in> (cstep r)\<^sup>*" by force
   with tseq have "(Fun f ts, Fun f ss) \<in> (cstep r)\<^sup>\<down>" by blast
 }
 with var show ?thesis  unfolding all_ctxt_closed_def by auto
qed

definition "quasi_reductive_order R S \<longleftrightarrow>
  trans S \<and> SN S \<and>
  ctxt_closed S \<and>
  (\<forall> l r cs (\<sigma> :: ('f, 'v) subst). ((l, r), cs) \<in> R \<longrightarrow> 
    (\<forall> i. i < length cs \<longrightarrow> (\<forall>j < i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S\<^sup>=) \<longrightarrow>
      (l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> (S \<union> {\<rhd>})\<^sup>+) \<and> 
        ((\<forall>j < length cs. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S\<^sup>=) \<longrightarrow> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S))"

definition "quasi_decreasing_order R S \<longleftrightarrow> 
  (trans S \<and> SN S \<and> 
  cstep R \<subseteq> S \<and> {\<rhd>} \<subseteq> S \<and> 
  (\<forall> l r cs (\<sigma> :: ('f, 'v) subst). ((l, r), cs) \<in> R \<longrightarrow> 
    (\<forall> i. i < length cs \<longrightarrow> (\<forall>j < i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (cstep R)\<^sup>* ) \<longrightarrow>
      (l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> S)))"

definition "quasi_reductive R \<longleftrightarrow> (\<exists>S. quasi_reductive_order R S)"

definition "quasi_decreasing R \<longleftrightarrow> (\<exists>S. quasi_decreasing_order R S)"

lemma quasi_reductive_order_cstep:
  assumes "quasi_reductive_order (R :: ('f, 'v) ctrs) S"
  shows "cstep R \<subseteq> S"
proof -
  from assms [unfolded quasi_reductive_order_def]
  have trans: "trans S"
  and mono: "ctxt_closed S"
  and qr: "\<And> l r cs (\<sigma> :: ('f,'v)subst).
             ((l, r), cs) \<in> R \<Longrightarrow>
             (\<forall>j<length cs. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=) \<Longrightarrow> (l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S" by auto
  show "cstep R \<subseteq> S"
  proof
    fix s t
    assume "(s,t) \<in> cstep R"
    then obtain n where "(s,t) \<in> cstep_n R n" unfolding cstep_def by auto
    thus "(s,t) \<in> S"
    proof (induct n arbitrary: s t)
      case (Suc n s t)
      from Suc(2)
      obtain l r cs C \<sigma> where lr: "((l,r),cs) \<in> R" and
        cs: "\<And> s\<^sub>i t\<^sub>i.  (s\<^sub>i,t\<^sub>i) \<in> set cs \<Longrightarrow> (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*"
        and s: "s = C \<langle> l \<cdot> \<sigma> \<rangle>" and t: "t = C \<langle> r \<cdot> \<sigma> \<rangle>" by (blast elim: cstep_n_SucE)
      note qr = qr[OF lr, of \<sigma>]
      have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> S"
      proof (rule qr[rule_format])
        fix j
        assume j: "j < length cs"
        obtain sj tj where csj: "cs ! j = (sj,tj)" by force
        from j have "(sj,tj) \<in> set cs" unfolding csj[symmetric] set_conv_nth by auto
        from cs[OF this] have steps: "(sj \<cdot> \<sigma>, tj \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*" .
        {
          fix s t
          assume "(s,t) \<in> (cstep_n R n)\<^sup>*"
          hence "(s,t) \<in> S^="
          proof (induct)
            case (step t u)
            from step(3) Suc(1)[OF step(2)] have "(s,u) \<in> S^= O S" by auto
            thus "(s,u) \<in> S^=" using trans[unfolded trans_def] by blast
          qed simp
        }
        from this[OF steps]
        show "(fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=" unfolding csj by simp
      qed 
      from ctxt_closedE[OF mono this, of C]
      show ?case unfolding s t .
    qed simp
  qed
qed

lemma quasi_reductive_order_quasi_decreasing_order: 
  assumes qro: "quasi_reductive_order (R :: ('f, 'v) ctrs) S"
  shows "quasi_decreasing_order R ((S \<union> {\<rhd>})\<^sup>+)"
proof -
  let ?S = "(S \<union> {\<rhd>})\<^sup>+"
  from quasi_reductive_order_cstep[OF qro] have "cstep R \<subseteq> S" by auto
  hence "cstep R \<subseteq> ?S" by auto
  from qro[unfolded quasi_reductive_order_def]
  have trans: "trans S" and SN: "SN S"
  and mono: "ctxt_closed S"
  and decr: "\<And> l r cs i (\<sigma> :: ('f,'v)subst).
             ((l, r), cs) \<in> R \<Longrightarrow>
              i  < length cs \<Longrightarrow>
               (\<forall>j<i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S\<^sup>=) \<Longrightarrow> (l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> ?S" 
    by auto
  from mono SN have "SN (S \<union> {\<rhd>})" by (metis SN_imp_SN_union_supt)
  hence "SN ?S" unfolding SN_trancl_SN_conv .
  have "trans ?S" by simp
  have "{\<rhd>} \<subseteq> ?S" by auto
  show ?thesis unfolding quasi_decreasing_order_def
  proof (intro conjI, rule \<open>trans ?S\<close>, rule \<open>SN ?S\<close>, rule \<open>cstep R \<subseteq> ?S\<close>, rule \<open>{\<rhd>} \<subseteq> ?S\<close>, intro allI impI)
    fix l r cs \<sigma> i
    assume lr: "((l,r), cs) \<in> R" and i: "i < length cs"
     and args: "\<forall>j<i. (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*"
    show "(l \<cdot> \<sigma>, fst (cs ! i) \<cdot> \<sigma>) \<in> ?S"
    proof (rule decr[OF lr i], intro allI impI)
      fix j
      assume "j < i"
      with args have "(fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> (cstep R)\<^sup>*" by auto
      also have "(cstep R)\<^sup>* \<subseteq> S^*" by (rule rtrancl_mono[OF \<open>cstep R \<subseteq> S\<close>])
      also have "S^* = (S^+)^=" by regexp
      also have "S^+ = S" using trans by auto
      finally show "(fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<in> S^=" .
    qed
  qed
qed

lemma quasi_reductive_quasi_decreasing: 
  "quasi_reductive R \<Longrightarrow> quasi_decreasing R"
  unfolding quasi_reductive_def quasi_decreasing_def
    using quasi_reductive_order_quasi_decreasing_order by blast

lemma quasi_decreasing_SN:
  assumes qd: "quasi_decreasing R"
  shows "SN (cstep R)"
proof -
  from qd[unfolded quasi_decreasing_def] obtain S where "quasi_decreasing_order R S" by blast
  from this[unfolded quasi_decreasing_order_def] have "SN S" and "cstep R \<subseteq> S" by auto
  thus ?thesis
    by (rule SN_subset)
qed

lemma quasi_reductive_SN:
  "quasi_reductive R \<Longrightarrow> SN (cstep R)"
  by (rule quasi_decreasing_SN[OF quasi_reductive_quasi_decreasing])

abbreviation clhs :: "('f, 'v) crule \<Rightarrow> ('f, 'v) term"
where
  "clhs \<rho> \<equiv> fst (fst \<rho>)"

abbreviation crhs :: "('f, 'v) crule \<Rightarrow> ('f, 'v) term"
where
  "crhs \<rho> \<equiv> snd (fst \<rho>)"

definition X_vars :: "('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'v set"
where
  "X_vars \<rho> i = vars_term (clhs \<rho>) \<union> \<Union>(vars_term ` rhss (set (take i (snd \<rho>))))"

definition Y_vars ::"('f, 'v) crule \<Rightarrow> nat \<Rightarrow> 'v set"
where
  "Y_vars \<rho> i =
    vars_term (crhs \<rho>) \<union> vars_term (snd (snd \<rho> ! i)) \<union> (vars_trs (set (drop i (snd \<rho>))))"

lemma X_vars_alt : 
  assumes "i \<le> length (snd \<rho>)"
  shows "X_vars \<rho> i = vars_term (clhs \<rho>) \<union> \<Union>((\<lambda>j. vars_term (snd (snd \<rho> ! j))) ` {j. j < i})"
proof -
  let ?L = "\<Union>(vars_term ` (rhss (set (List.take i (snd \<rho>)))))"
  let ?R = "\<Union>((\<lambda>j. vars_term (snd (snd \<rho> ! j))) ` {j. j < i})"
  { fix x
   assume "x \<in> ?L" 
   hence "\<exists>j. j < i \<and> x \<in> vars_term (snd ((snd \<rho>) ! j))" unfolding set_conv_nth using nth_take[of _ i "snd \<rho>"] by fastforce
   then obtain j where j:"j < i" and x:"x \<in> vars_term (snd (snd \<rho> ! j))" by blast
   hence "x \<in> ?R"  by blast
  } hence *: "?L \<subseteq> ?R" by auto
  { fix x
   assume "x \<in> ?R"
   hence "\<exists>j. j < i \<and> x \<in> vars_term (snd ((snd \<rho>) ! j))" by blast
   then obtain j where  j:"j < i" and x:"x \<in> vars_term (snd (snd \<rho> ! j))" by blast
   hence "x \<in> vars_term (snd (take i (snd \<rho>) ! j))"  using nth_take[OF j(1)] by force
   with j(1)  have "x \<in> ?L" unfolding set_conv_nth length_take min_absorb2[OF assms] by blast
  } hence "?R \<subseteq> ?L" by auto
  with * have *:"?L = ?R" by blast
  show ?thesis unfolding X_vars_def * by blast
qed

definition type3 :: "('f, 'v) ctrs \<Rightarrow> bool"
where
  "type3 R \<longleftrightarrow> (\<forall> \<rho> \<in> R. vars_term (crhs \<rho>) \<subseteq> (vars_term (clhs \<rho>) \<union> vars_trs (set (snd \<rho>))))"

definition dctrs :: "('f, 'v) ctrs \<Rightarrow> bool"
where
  "dctrs R \<longleftrightarrow> (\<forall> \<rho> \<in> R. \<forall> i < length (snd \<rho>). vars_term (fst (snd \<rho> ! i)) \<subseteq> X_vars \<rho> i)"

definition wf_ctrs :: "('f, 'v) ctrs \<Rightarrow> bool"
where
  "wf_ctrs R \<longleftrightarrow> (\<forall>((l, r), cs) \<in> R. is_Fun l) \<and> type3 R \<and> dctrs R"

definition funs_crule :: "('f, 'v) crule \<Rightarrow> 'f set"
where
  "funs_crule \<rho> = funs_rule (fst \<rho>) \<union> funs_trs (set (snd \<rho>))"

definition funas_crule :: "('f, 'v) crule \<Rightarrow> ('f \<times> nat) set"
where
  "funas_crule \<rho> = funas_rule (fst \<rho>) \<union> funas_trs (set (snd \<rho>))"

definition funs_ctrs :: "('f, 'v) ctrs \<Rightarrow> 'f set"
where
  "funs_ctrs R = \<Union>(funs_crule ` R)"

definition funas_ctrs :: "('f, 'v) ctrs \<Rightarrow> ('f \<times> nat) set"
where
  "funas_ctrs R = \<Union>(funas_crule ` R)"

definition vars_crule :: "('f, 'v) crule \<Rightarrow> 'v set"
where
  "vars_crule \<rho> = vars_rule (fst \<rho>) \<union> vars_trs (set (snd \<rho>))"

definition vars_ctrs :: "('f, 'v) ctrs \<Rightarrow> 'v set"
where
  "vars_ctrs R = \<Union>(vars_crule ` R)"

text \<open>The underlying unconditional TRS\<close>
definition Ru :: "('f, 'v) ctrs \<Rightarrow> ('f, 'v) trs"
where
  "Ru R = fst ` R"

lemma defined_R_imp_Ru: "root l = Some fn \<Longrightarrow> ((l, r), cs) \<in> R \<Longrightarrow> defined (Ru R) fn"
  by (force simp: Ru_def defined_def)
  
lemma cstep_imp_Ru_step: "(cstep R) \<subseteq> (rstep (Ru R))"
proof
 fix s t
 assume "(s,t) \<in> cstep R"
 with cstep_iff cstep_n.simps(1) obtain n where n:"(s,t) \<in> cstep_n R (Suc n)"
  by (metis empty_iff not0_implies_Suc)
 hence "\<exists>C l r \<sigma> cs. (s,t) = (C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<and> ((l,r),cs) \<in> R" unfolding Ru_def cstep_n.simps(2) by fast
 hence *:"\<exists>C l r \<sigma>. (s,t) = (C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>) \<and> (l,r) \<in> Ru R" unfolding Ru_def by force
 with rule rstep_r_p_s_imp_rstep * rstep_r_p_s_def show "(s,t) \<in> rstep (Ru R)" by fast
qed

lemma Ru_NF_imp_R_NF: "t \<in> NF (rstep (Ru R)) \<Longrightarrow> t \<in> NF (cstep R)"
using assms NF_anti_mono [OF cstep_imp_Ru_step] by blast

lemma funas_ctrs_cond:
 assumes rho:"\<rho> \<in> R" and i:"i < length (snd \<rho>)" shows "funas_rule (snd \<rho> ! i) \<subseteq> funas_ctrs R"
   using assms funas_crule_def unfolding funas_ctrs_def funas_trs_def set_conv_nth by blast

lemma wf_ctrs_steps_preserves_funas:
 assumes wf: "wf_ctrs R" and s: "funas_term s \<subseteq> F" and R: "funas_ctrs R \<subseteq> F"
   and st:"(s, t) \<in> (cstep R)\<^sup>*"
 shows "funas_term t \<subseteq> F"
proof- 
 from wf[unfolded wf_ctrs_def] have t3:"type3 R" and dctrs:"dctrs R" by auto
 let ?rfuns = "\<lambda> v. funas_term v \<subseteq> F"
 from rtrancl_power csteps_imp_csteps_n[OF st]  obtain n k where st':"(s,t) \<in> (cstep_n R n) ^^ k" by blast
 let ?P = "\<lambda> (n,k). \<forall> s t. ?rfuns s \<longrightarrow> (s,t) \<in> (cstep_n R n) ^^ k \<longrightarrow> ?rfuns t"
 have "?P (n,k)" proof(induct rule: wf_induct[OF wf_measures, of "[fst, snd]" ?P])
  case (1 nk) note ind = this
   obtain n k where nk: "nk = (n,k)" by force
   { fix s t
     assume s:"?rfuns s" and st:"(s, t) \<in> (cstep_n R n) ^^ k"
     hence "?rfuns t" proof(cases k)
      case 0
       with s st show ?thesis by simp
      next
      case (Suc m)
       from relpow_Suc_E[OF st[unfolded Suc]] obtain u where u:"(s,u) \<in> (cstep_n R n) ^^ m" "(u,t) \<in> cstep_n R n" by blast 
       have "((n,m), nk) \<in> measures [fst, snd]" unfolding nk Suc by auto
       from ind[rule_format, OF this, unfolded split, rule_format, OF s u(1)] have fu:"?rfuns u" by auto
       show ?thesis proof (cases n)
        case 0
         from u(2)[unfolded 0 cstep_n.simps] show ?thesis by auto
        next
        case (Suc n)
          from u(2)[unfolded Suc cstep_n.simps] obtain C l r \<sigma> cs where step:"(u,t) = (C\<langle>l \<cdot> \<sigma>\<rangle>, C\<langle>r \<cdot> \<sigma>\<rangle>)"
           "((l, r), cs) \<in> R" "\<forall>(s\<^sub>i, t\<^sub>i)\<in>set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*" by force
          from this(1) fu have fl:"?rfuns (l \<cdot> \<sigma>)" by auto
          with supteq_imp_funas_term_subset supteq_Var
            have fx:"\<And>x. x \<in> vars_term l \<Longrightarrow> ?rfuns (\<sigma> x)" by fastforce
          { fix i
            assume i:"i < length cs"
            let ?stm = " \<lambda> m. (fst (cs ! m) \<cdot> \<sigma>, snd (cs ! m) \<cdot> \<sigma>)"
            from i have "funas_rule (?stm i) \<subseteq> F" proof (induct i rule:nat_less_induct)
             fix i
             assume ind2:"\<forall>m<i. (m < length cs \<longrightarrow> funas_rule (?stm m) \<subseteq> F)" and i:"i < length cs"
             from dctrs[unfolded dctrs_def, rule_format, OF step(2), of i] i have 
              vs:"vars_term (fst (cs ! i)) \<subseteq> vars_term l \<union> X_vars ((l, r), cs) i" by auto
             { fix j x
               assume "x \<in> \<Union>(vars_term ` rhss (set (take i cs)))" 
               hence "\<exists>j. j < i \<and> x \<in> vars_term (snd (cs ! j))" unfolding set_conv_nth using nth_take[of _ i cs] by fastforce
               then obtain j where j:"j < i" and x:"x \<in> vars_term (snd (cs ! j))" by blast
               from ind2[rule_format, OF j] i j have "funas_term (snd (cs ! j) \<cdot> \<sigma>) \<subseteq> F" unfolding funas_rule_def by auto
               with x supteq_imp_funas_term_subset supteq_Var have "?rfuns (\<sigma> x)" by fastforce
             } note vt = this
             { fix x assume "x \<in> vars_term (fst (cs ! i))"
               with vs[unfolded X_vars_def] fx vt have "?rfuns (\<sigma> x)" by auto
             }
             with funas_ctrs_cond[OF step(2), unfolded snd_conv, OF i] R have lhs:"?rfuns (fst (cs ! i) \<cdot> \<sigma>)" 
              unfolding funas_term_subst funas_rule_def by blast
             from i have "(fst (cs ! i), snd (cs ! i)) \<in> set cs" by auto
             from  step(3)[rule_format, OF this, unfolded split] obtain l where 
              st:"(fst (cs ! i) \<cdot> \<sigma>, snd (cs ! i) \<cdot> \<sigma>) \<in> (cstep_n R n) ^^ l" unfolding rtrancl_power by blast
             have "((n,l), nk) \<in> measures [fst, snd]" unfolding nk Suc by auto
             from ind[rule_format, OF this, unfolded split, rule_format, OF lhs st]
               have rhs:"funas_term (snd (cs ! i) \<cdot> \<sigma>) \<subseteq> F" by auto
             from lhs rhs show "funas_rule (?stm i) \<subseteq> F" unfolding funas_rule_def by auto
            qed
          } note aux = this
          from t3[unfolded type3_def, rule_format, OF step(2)] 
           have vr:"vars_term r \<subseteq> vars_term l \<union> vars_trs (set cs )" by auto
           from funas_rule_def[of "(l,r)"] have "funas_term r \<subseteq> funas_crule ((l, r), cs)" unfolding funas_crule_def by auto
           with  step(2) have fr:"funas_term r \<subseteq> funas_ctrs R" unfolding funas_ctrs_def by auto
          { fix j x
            assume j:"j < length cs" and "x \<in> vars_rule (cs ! j)"
            with supteq_Var have "fst (cs ! j) \<unrhd> Var x \<or> snd (cs ! j) \<unrhd> Var x" unfolding vars_rule_def by fastforce
            hence x:"fst (cs ! j) \<cdot> \<sigma> \<unrhd> Var x \<cdot> \<sigma> \<or> snd (cs ! j) \<cdot> \<sigma> \<unrhd> Var x \<cdot> \<sigma>" by auto
            from aux j have "funas_rule (fst (cs ! j) \<cdot> \<sigma>, snd (cs ! j) \<cdot> \<sigma>) \<subseteq> F" unfolding funas_rule_def by auto
            with x supteq_imp_funas_term_subset  have "?rfuns (\<sigma> x)" unfolding funas_rule_def by force
          }
          with vr[unfolded set_conv_nth vars_trs_def] fx have "\<And>x. x \<in> vars_term r \<Longrightarrow> ?rfuns (\<sigma> x)" by blast
          with step(2) fr R have lhs:"?rfuns (r \<cdot> \<sigma>)" unfolding funas_term_subst funas_rule_def by blast
          from step(1) have t:"t = C\<langle>r \<cdot> \<sigma>\<rangle>" and u:"u = C\<langle>l \<cdot> \<sigma>\<rangle>" by (fast,fast)
          from lhs fu show "?rfuns t" unfolding t u by auto
         qed
      qed
   }
   thus ?case unfolding nk split by blast
 qed
 from this[unfolded split, rule_format, OF s st'] show ?thesis by auto
qed

lemma funs_crule_funas_crule:
  "funs_crule r = fst ` funas_crule r" 
  unfolding funs_crule_def funas_crule_def  funs_trs_funas_trs funs_rule_funas_rule by blast

lemma funs_ctrs_funas_ctrs:
  "funs_ctrs R = fst ` funas_ctrs R" 
proof -
 from funs_crule_funas_crule have "\<Union>(funs_crule ` R) = \<Union>(((\<lambda>x. fst ` x) \<circ> funas_crule) ` R)" by (metis comp_apply)
 thus ?thesis unfolding funs_ctrs_def funas_ctrs_def image_Union by force
qed

lemma wf_ctrs_steps_preserves_funs:
  assumes wf: "wf_ctrs R" and s: "funs_term s \<subseteq> F"
    and R: "funs_ctrs R \<subseteq> F" and st: "(s, t) \<in> (cstep R)\<^sup>*"
  shows "funs_term t \<subseteq> F"
proof -
  from s R wf_ctrs_steps_preserves_funas[OF wf _ _ st, of "F \<times> UNIV"]
  show ?thesis unfolding funs_term_funas_term funs_ctrs_funas_ctrs by fastforce
qed

lemma cstep_map_funs_term:
  assumes R: "\<And> f. f \<in> funs_ctrs R \<Longrightarrow> h f = f" and "(s, t) \<in> cstep R"
  shows "(map_funs_term h s, map_funs_term h t) \<in> cstep R"
proof -
  let ?h = "map_funs_term h"
  from assms[unfolded cstep_def] obtain n where "(s,t) \<in> cstep_n R n" by auto
  hence "(?h s, ?h t) \<in> cstep_n R n"
  proof (induct n arbitrary: s t)
    case (Suc n)
    from cstep_n_SucE[OF Suc(2)]
      obtain C l r \<sigma> cs
      where rule: "((l,r), cs) \<in> R" and cs: "\<forall>(s\<^sub>i, t\<^sub>i)\<in>set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*" 
      and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" . 
    let ?\<sigma> = "map_funs_subst h \<sigma>"
    note funs_defs = funs_ctrs_def funs_crule_def[abs_def] funs_rule_def[abs_def] funs_trs_def
    from rule have lr: "funs_term l \<union> funs_term r \<subseteq> funs_ctrs R" unfolding funs_defs
      by auto
    have hl: "?h l = l"
      by (rule funs_term_map_funs_term_id[OF R], insert lr, auto)
    have hr: "?h r = r"
      by (rule funs_term_map_funs_term_id[OF R], insert lr, auto)
    show ?case unfolding s t
      unfolding map_funs_subst_distrib map_funs_term_ctxt_distrib hl hr
    proof (rule cstep_n_SucI[OF rule _ refl refl], rule)
      fix s\<^sub>i t\<^sub>i
      assume in_cs: "(s\<^sub>i,t\<^sub>i) \<in> set cs"
      from in_cs rule have st: "funs_term s\<^sub>i \<union> funs_term t\<^sub>i \<subseteq> funs_ctrs R" unfolding funs_defs by force
      have hs: "?h s\<^sub>i = s\<^sub>i"
        by (rule funs_term_map_funs_term_id[OF R], insert st, auto)
      have ht: "?h t\<^sub>i = t\<^sub>i"
        by (rule funs_term_map_funs_term_id[OF R], insert st, auto)
      from in_cs cs have steps: "(s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R n)\<^sup>*" by auto
      have "(?h (s\<^sub>i \<cdot> \<sigma>), ?h (t\<^sub>i \<cdot> \<sigma>)) \<in> (cstep_n R n)\<^sup>*"
        by (rule rtrancl_map[OF Suc(1) steps])
      thus "(s\<^sub>i \<cdot> ?\<sigma>, t\<^sub>i \<cdot> ?\<sigma>) \<in> (cstep_n R n)\<^sup>*" unfolding map_funs_subst_distrib hs ht .
    qed
  qed simp
  thus ?thesis unfolding cstep_def by auto
qed

definition normalized :: "('f, 'v) ctrs \<Rightarrow> ('f, 'v) subst \<Rightarrow> bool"
where
  "normalized R \<sigma> \<longleftrightarrow> (\<forall>x. \<sigma> x \<in> NF (cstep R))"

lemma empty_subst_normalized:
  assumes "\<forall>((l, r), cs) \<in> R. is_Fun l"
  shows "normalized R Var"
  by (simp add: Var_NF_cstep assms normalized_def)
  
definition "strongly_irreducible R t \<longleftrightarrow> (\<forall>\<sigma>. normalized R \<sigma> \<longrightarrow> t \<cdot> \<sigma> \<in> NF (cstep R))"

lemma strongly_irreducible_I [Pure.intro?]:
  assumes "\<And>\<sigma>. normalized R \<sigma> \<Longrightarrow> t \<cdot> \<sigma> \<in> NF (cstep R)"
  shows "strongly_irreducible R t"
using assms by (auto simp add: strongly_irreducible_def)

lemma Var_strongly_irreducible_cstep:
  "\<forall> ((l, r), cs) \<in> R. is_Fun l \<Longrightarrow> strongly_irreducible R (Var x)"
by (rule strongly_irreducible_I) (simp add: normalized_def)

lemma constructor_term_imp_strongly_irreducible:
  fixes t R
  assumes varcond: "\<forall>((l, r), cs) \<in> R. is_Fun l"
    and constr: "funas_term t \<subseteq> funas_ctrs R - { f. defined (Ru R) f }"
  shows "strongly_irreducible R t"
using constr
proof (induction t)
  case (Var x)
  show ?case using varcond Var_strongly_irreducible_cstep[of R x] by simp
next
  case (Fun f ts)
  show ?case
  proof
    fix \<sigma>
    assume norm: "normalized R \<sigma>"
    show "Fun f ts \<cdot> \<sigma> \<in> NF (cstep R)"
    proof
      fix s
      show "(Fun f ts \<cdot> \<sigma>, s) \<notin> cstep R"
      proof
        assume "(Fun f ts \<cdot> \<sigma>, s) \<in> cstep R"
        with cstepE [of "Fun f ts \<cdot> \<sigma>" s "R"] obtain C \<tau> l r cs
        where "((l, r), cs) \<in> R" and "Fun f ts \<cdot> \<sigma> = C\<langle>l \<cdot> \<tau>\<rangle>" and "s = C\<langle>r \<cdot> \<tau>\<rangle>"
        and "\<forall>(s\<^sub>i, t\<^sub>i)\<in>set cs. (s\<^sub>i \<cdot> \<tau>, t\<^sub>i \<cdot> \<tau>) \<in> (cstep R)\<^sup>*" by auto
        from varcond \<open>((l, r), cs) \<in> R\<close> Ru_def have "is_Fun l" by fastforce
        show "False"
        proof (cases C)
          case (Hole)
          with \<open>Fun f ts \<cdot> \<sigma> = C\<langle>l \<cdot> \<tau>\<rangle>\<close> have "Fun f ts \<cdot> \<sigma> = l \<cdot> \<tau>" by auto
          with \<open>is_Fun l\<close> \<open>((l, r), cs) \<in> R\<close>
            have "(f, length ts) \<in> {f. defined (Ru R) f}"
            using map_eq_imp_length_eq [of "\<lambda>t. t \<cdot> \<sigma>" ts]
            by (cases l) (force intro: defined_R_imp_Ru)+
          with Fun.prems show "False" by auto
        next
          case (More g lts D rts)
          {
            fix u
            assume "u \<in> set ts"
            then have "funas_term u \<subseteq> funas_term (Fun f ts)" by auto
            with Fun.prems have "funas_term u \<subseteq> funas_ctrs R - {f. defined (Ru R) f}" by auto
          }
          note * = this
          {
            fix u
            assume "u \<in> set ts"
            from * [OF this] and Fun.IH[OF this]
            have "strongly_irreducible R u" by auto
          }
          note IH = this
          from More \<open>Fun f ts \<cdot> \<sigma> = C\<langle>l \<cdot> \<tau>\<rangle>\<close> have "map (\<lambda>t. t \<cdot> \<sigma>) ts = lts @ D\<langle>l \<cdot> \<tau>\<rangle> # rts" by simp
          then obtain u where u: "u \<in> set ts" and "u \<cdot> \<sigma> = D\<langle>l \<cdot> \<tau>\<rangle>"
            by (induct ts arbitrary: lts rts) (auto simp: Cons_eq_append_conv)
          then have "u \<cdot> \<sigma> \<unrhd> l \<cdot> \<tau>" by simp
          then show "False"
          proof (cases rule: supteq_subst_cases)
            note supteq_imp_funas_term_subset [simp del]
            case (in_term v)
            with \<open>is_Fun l\<close> and * [OF u] and defined_R_imp_Ru[OF _ \<open>((l, r), cs) \<in> R\<close>]
            show ?thesis
            using supteq_imp_funas_term_subset [OF \<open>u \<unrhd> v\<close>] and map_eq_imp_length_eq
            by (cases l; cases v) (fastforce)+
          next
            have "(l \<cdot> \<tau>, r \<cdot> \<tau>) \<in> cstep R" by (rule cstepI [of _ _ _ _ _ _ \<box>], fact+) (simp_all)
            moreover
            case (in_subst x)
            ultimately show ?thesis using norm NF_cstep_subterm[of "\<sigma> x" R "l \<cdot> \<tau>"]
              by (auto simp: normalized_def)
          qed
        qed
      qed
    qed
  qed
qed

lemma strongly_irreducible_imp_NF:
  assumes vc: "\<forall>((l, r), cs) \<in> R. is_Fun l"
    and si: "strongly_irreducible R t"
  shows "t \<in> NF (cstep R)"
by (metis empty_subst_normalized si strongly_irreducible_def subst_apply_term_empty vc)

lemma constructor_term_NF_cstep:
  assumes varcond: "\<forall>((l, r), cs) \<in> R. is_Fun l"
    and  constr: "funas_term t \<subseteq> funas_ctrs R - { f. defined (Ru R) f}"
  shows "t \<in> NF (cstep R)"
by (auto simp: assms constructor_term_imp_strongly_irreducible strongly_irreducible_imp_NF)

lemma ground_Ru_NF_imp_strongly_irreducible:
  assumes varcond: "\<forall>((l, r), cs) \<in> R. is_Fun l"
    and ground: "ground t"
    and ru_nf: "t \<in> NF (rstep (Ru R))"
  shows "strongly_irreducible R t"
using ground ru_nf
proof (induction t)
  case (Fun f ts)
  from Fun.prems(1) have "\<forall>\<sigma>. Fun f ts \<cdot> \<sigma> = Fun f ts"
  using ground_subst_apply by blast
  then show ?case
    using Fun.prems(2) Ru_NF_imp_R_NF [of "Fun f ts" R] strongly_irreducible_I[of R "Fun f ts"]
    by fastforce
qed simp

context
  fixes R :: "('f, 'v) ctrs"
  assumes vc: "\<forall>((l, r), cs) \<in> R. is_Fun l"
begin

lemma syntactic_det_imp_strong_det:
  assumes "(\<forall> ((l, r), cs) \<in> R. \<forall> (s\<^sub>i, t\<^sub>i) \<in> set cs.
    ((funas_term t\<^sub>i \<subseteq> ((funas_ctrs R) - { f. defined (Ru R) f }) \<or>
    (ground t\<^sub>i \<and> t\<^sub>i \<in> NF (rstep (Ru R))))))"
  shows str_det: "(\<forall> ((l, r), cs) \<in> R. \<forall> (s\<^sub>i, t\<^sub>i) \<in> set cs. strongly_irreducible R t\<^sub>i)"
proof -
  {
    fix l r cs s\<^sub>i t\<^sub>i
    assume "((l, r), cs) \<in> R" and "(s\<^sub>i, t\<^sub>i) \<in> set cs"
    moreover
    {
      assume *: "funas_term t\<^sub>i \<subseteq> funas_ctrs R - { f. defined (Ru R) f }"
      from vc have **: "\<forall>((l, r), cs) \<in> R. is_Fun l" by auto
      have "strongly_irreducible R t\<^sub>i"
        using constructor_term_imp_strongly_irreducible [OF ** *] by auto
    }
    moreover
    {
      assume "ground t\<^sub>i \<and> t\<^sub>i \<in> NF (rstep (Ru R))"
      then have "strongly_irreducible R t\<^sub>i"
        using ground_Ru_NF_imp_strongly_irreducible
        by (simp add: Ru_NF_imp_R_NF ground_subst_apply strongly_irreducible_I)
    }
    ultimately
    have "strongly_irreducible R t\<^sub>i" using assms by fastforce
  }
  then show ?thesis by force
qed

end

lemma Var_cstep_n [dest]:
  assumes "(Var x, t) \<in> cstep_n R n"
    and "\<forall>((l, r), cs) \<in> R. is_Fun l"
  shows "False"
using assms by (auto simp: elim!: cstep_nE) (case_tac C; case_tac l; force)

lemma Var_rtrancl_cstep_n [dest]:
  assumes "(Var x, t) \<in> (cstep_n R n)\<^sup>*"
    and "\<forall>((l, r), cs) \<in> R. is_Fun l"
  shows "t = Var x"
using assms by (induct) auto


subsection \<open>TRS for level @{term n}.\<close>

fun trs_n :: "('f, 'v) ctrs \<Rightarrow> nat \<Rightarrow> ('f, 'v) trs"
where
  "trs_n R 0 = {}" |
  "trs_n R (Suc n) =
    {(l \<cdot> \<sigma>, r \<cdot> \<sigma>) | l r \<sigma> cs.
      ((l, r), cs) \<in> R \<and> (\<forall> (s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (rstep (trs_n R n))\<^sup>*)}"

lemma trs_n_SucI:
  assumes "((l, r), cs) \<in> R"
    and "\<And>s\<^sub>i t\<^sub>i. (s\<^sub>i, t\<^sub>i) \<in> set cs \<Longrightarrow> (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (rstep (trs_n R n))\<^sup>*"
  shows "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> trs_n R (Suc n)"
  using assms by (simp) blast

lemma trs_nE:
  assumes "(s, t) \<in> trs_n R n"
  obtains l r \<sigma> cs n' where "n = Suc n'" and "s = l \<cdot> \<sigma>" and "t = r \<cdot> \<sigma>" and "((l, r), cs) \<in> R"
    and "\<forall> (s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (rstep (trs_n R n'))\<^sup>*"
  using assms by (cases n; simp) blast

lemma trs_n_SucE:
  assumes "(s, t) \<in> trs_n R (Suc n)"
  obtains l r \<sigma> cs where "s = l \<cdot> \<sigma>" and "t = r \<cdot> \<sigma>" and "((l, r), cs) \<in> R"
    and "\<forall> (s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (rstep (trs_n R n))\<^sup>*"
  using assms by (simp) blast

declare trs_n.simps(2) [simp del]

lemma trs_n_subst:
  assumes "(l, r) \<in> trs_n R n"
  shows "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> trs_n R n"
using assms
proof (induction n)
  case (Suc n)
  from Suc.prems obtain l' r' \<sigma>' cs
    where [simp]: "l = l' \<cdot> \<sigma>'" "r = r' \<cdot> \<sigma>'" and "((l', r'), cs) \<in> R"
    and "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>', t\<^sub>i \<cdot> \<sigma>') \<in> (rstep (trs_n R n))\<^sup>*"
    by (auto elim: trs_n_SucE)
  moreover then have "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs.
    (s\<^sub>i \<cdot> (\<sigma>' \<circ>\<^sub>s \<sigma>), t\<^sub>i \<cdot> (\<sigma>' \<circ>\<^sub>s \<sigma>)) \<in> (rstep (trs_n R n))\<^sup>*" by (auto intro: rsteps_closed_subst)
  ultimately have "(l' \<cdot> (\<sigma>' \<circ>\<^sub>s \<sigma>), r' \<cdot> (\<sigma>' \<circ>\<^sub>s \<sigma>)) \<in> trs_n R (Suc n)"
    by (blast intro: trs_n_SucI)
  then show ?case by simp
qed simp

lemma no_step_from_constructor:
  assumes vc: "\<forall>((l, r), cs) \<in> R. is_Fun l"
    and "funas_term t \<subseteq> funas_ctrs R - { f. defined (Ru R) f }"
    and fp: "p \<in> funposs t"
  shows "(t \<cdot> \<sigma> |_ p, u) \<notin> trs_n R n"
proof
  assume "(t \<cdot> \<sigma> |_ p, u) \<in> trs_n R n"
  from trs_nE [OF this] obtain l r \<tau> cs where *: "((l, r), cs) \<in> R" and **: "t \<cdot> \<sigma> |_ p = l \<cdot> \<tau>"
    by metis
  then obtain f ts where [simp]: "l = Fun f ts" using vc by (cases l, auto)
  have "defined (Ru R) (f, length ts)" using * by (intro defined_R_imp_Ru) (auto)
  moreover from fp ** have "(f, length ts) \<in> funas_term t"
  proof -
    from funposs_fun_conv [OF fp] obtain g ss where ***: "t |_ p = Fun g ss" by blast
    with funposs_imp_poss [OF fp] have "t \<cdot> \<sigma> |_ p = Fun g (map (\<lambda>x. x \<cdot> \<sigma>) ss)" by auto
    then have [simp]: "g = f" "length ts = length ss" using ** map_eq_imp_length_eq by (force)+
    from *** funposs_imp_poss [OF fp] show ?thesis by (auto simp add: funas_term_poss_conv)   
  qed
  ultimately show "False" using assms(2) by blast
qed

lemma cstep_n_rstep_trs_n_conv:
  "cstep_n R n = rstep (trs_n R n)"
proof (induction n)
  case (Suc n)
  { fix s t
    assume "(s, t) \<in> cstep_n R (Suc n)"
    then have "(s, t) \<in> rstep (trs_n R (Suc n))"
      by (auto elim!: cstep_n_SucE simp: Suc split: prod.splits)
         (metis rstep.ctxt rstep.id trs_n_SucI) }
  moreover
  { fix s t
    assume "(s, t) \<in> rstep (trs_n R (Suc n))"
    then obtain u v C \<sigma> where "(u, v) \<in> trs_n R (Suc n)"
      and [simp]: "s = C\<langle>u \<cdot> \<sigma>\<rangle>""t = C\<langle>v \<cdot> \<sigma>\<rangle>" ..
    then obtain l r cs \<tau> where "((l, r), cs) \<in> R"
      and [simp]: "u = l \<cdot> \<tau>" "v = r \<cdot> \<tau>"
      and "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<tau>, t\<^sub>i \<cdot> \<tau>) \<in> (rstep (trs_n R n))\<^sup>*"
      by (auto elim: trs_n_SucE)
    moreover then have "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> (\<tau> \<circ>\<^sub>s \<sigma>), t\<^sub>i \<cdot> (\<tau> \<circ>\<^sub>s \<sigma>)) \<in> (cstep_n R n)\<^sup>*"
      by (auto intro: rsteps_closed_subst simp: Suc)
    ultimately have "(C\<langle>l \<cdot> (\<tau> \<circ>\<^sub>s \<sigma>)\<rangle>, C\<langle>r \<cdot> (\<tau> \<circ>\<^sub>s \<sigma>)\<rangle>) \<in> cstep_n R (Suc n)" by (blast intro: cstep_n_SucI)
    then have "(s, t) \<in> cstep_n R (Suc n)" by simp }
  ultimately show ?case by auto
qed simp

lemma trs_n_Suc_mono:
  "trs_n R n \<subseteq> trs_n R (Suc n)"
proof (induct n)
  case (Suc n)
  then have "rstep (trs_n R n) \<subseteq> rstep (trs_n R (Suc n))" by (metis rstep_mono)
  from rtrancl_mono [OF this]
    show ?case by (auto elim!: trs_n_SucE intro!: trs_n_SucI)
qed simp

lemma rrstep_trs_n [simp]:
  "rrstep (trs_n R n) = trs_n R n"
  by (auto simp: CS_rrstep_conv [symmetric]
    elim: subst_closure.cases intro: subst_closureI2 [of _ _ _ _ Var] trs_n_subst)

lemma Var_trs_n [dest]:
  assumes "(Var x, t) \<in> trs_n R n"
    and "\<forall>((l, r), cs) \<in> R. is_Fun l"
  shows "False"
using assms by (induct n) (force elim!: trs_n_SucE split: prod.splits)+

lemma linear_term_cstep_n_cases':
  assumes "linear_term s"
    and cstep: "(s \<cdot> \<sigma>, u) \<in> cstep_n R n"
  shows "(\<exists>\<tau>. subst_domain \<tau> \<subseteq> vars_term s \<and> s \<cdot> \<tau> = u \<and> (\<forall>x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> (cstep_n R n)\<^sup>*)) \<or>
    (\<exists>p \<in> funposs s. (s \<cdot> \<sigma> |_ p, u |_ p) \<in> trs_n R n)"
proof -
  from cstep [THEN cstep_nE] obtain C l r \<sigma>' cs k
    where rule: "((l, r), cs) \<in> R" and [simp]: "n = Suc k"
    and conds: "conds_sat R k cs \<sigma>'"
    and s_\<sigma>: "s \<cdot> \<sigma> = C\<langle>l \<cdot> \<sigma>'\<rangle>" and u: "u = C\<langle>r \<cdot> \<sigma>'\<rangle>"
    unfolding conds_sat_iff by metis (*only metis works; investigate!*)
  def p \<equiv> "hole_pos C"
  have *: "p \<in> poss (s \<cdot> \<sigma>)" by (simp add: assms s_\<sigma> p_def)
  show ?thesis
  proof (cases "p \<in> funposs s")
    case False
    with * obtain q\<^sub>1 q\<^sub>2 x
      where p: "p = q\<^sub>1 <#> q\<^sub>2" and q\<^sub>1: "q\<^sub>1 \<in> poss s"
      and lq\<^sub>1: "s |_ q\<^sub>1 = Var x" and q\<^sub>2: "q\<^sub>2 \<in> poss (\<sigma> x)"
        by (rule poss_subst_apply_term)
    moreover
    have [simp]: "s \<cdot> \<sigma> |_ p = l \<cdot> \<sigma>'" using assms p_def s_\<sigma> by auto 
    ultimately
    have [simp]: "\<sigma> x |_ q\<^sub>2 = l \<cdot> \<sigma>'" by simp
    
    def \<tau> \<equiv> "\<lambda>y. if y = x then replace_at (\<sigma> x) q\<^sub>2 (r \<cdot> \<sigma>') else (\<sigma> |s vars_term s) y"
    have \<tau>_x: "\<tau> x = replace_at (\<sigma> x) q\<^sub>2 (r \<cdot> \<sigma>')" by (simp add: \<tau>_def)
    have [simp]: "\<And>y. y \<in> vars_term s \<Longrightarrow> y \<noteq> x \<Longrightarrow> \<tau> y = \<sigma> y" by (simp add: \<tau>_def)
    have "(\<sigma> x, \<tau> x) \<in> cstep_n R n"
    proof -
      let ?C = "ctxt_of_pos_term q\<^sub>2 (\<sigma> x)"
      have "(?C\<langle>l \<cdot> \<sigma>'\<rangle>, ?C\<langle>r \<cdot> \<sigma>'\<rangle>) \<in> cstep_n R n"
        using conds by (auto intro!: cstep_n_SucI rule simp: conds_sat_iff)
      then show ?thesis using q\<^sub>2 by (simp add: \<tau>_def replace_at_ident)
    qed
    then have "\<forall>x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> (cstep_n R n)\<^sup>*" by (auto simp: \<tau>_def)
    moreover have "subst_domain \<tau> \<subseteq> vars_term s"
      using lq\<^sub>1 and vars_term_subt_at [OF q\<^sub>1]
      by (auto simp: subst_domain_def \<tau>_def subst_restrict_def)
    moreover have "s \<cdot> \<tau> = u"
    proof -
      have [simp]: "ctxt_of_pos_term p (s \<cdot> \<sigma>) = C" by (simp add: assms s_\<sigma> p_def)
      have "s \<cdot> \<sigma> |_ q\<^sub>1 = \<sigma> x" by (simp add: lq\<^sub>1 q\<^sub>1)
      from linear_term_replace_in_subst [OF \<open>linear_term s\<close> q\<^sub>1 lq\<^sub>1, of \<sigma> \<tau>, OF _ \<tau>_x, folded ctxt_compose_apply_distrib]
        and q\<^sub>1 and ctxt_of_pos_term_append [of q\<^sub>1 "s \<cdot> \<sigma>" q\<^sub>2, folded p, unfolded this, symmetric]
        show "s \<cdot> \<tau> = u" by (simp add: u)
    qed
    ultimately show ?thesis by blast
  next
    case True
    moreover have "(s \<cdot> \<sigma> |_ p, u |_ p) \<in> trs_n R (Suc k)"
      using conds
      by (auto simp: conds_sat_iff p_def s_\<sigma> u
               intro: trs_n_SucI [OF rule] simp: cstep_n_rstep_trs_n_conv)
    ultimately show ?thesis by auto
  qed
qed

lemma linear_term_cstep_n_cases [consumes 2]:
  assumes "linear_term s"
    and "(s \<cdot> \<sigma>, u) \<in> cstep_n R n"
  obtains (varposs) \<tau> where "subst_domain \<tau> \<subseteq> vars_term s" and "s \<cdot> \<tau> = u"
    and "\<forall>x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> (cstep_n R n)\<^sup>*"
  | (funposs) p where "p \<in> funposs s" and "(s \<cdot> \<sigma> |_ p, u |_ p) \<in> trs_n R n"
using linear_term_cstep_n_cases' [OF assms] by blast

lemma linear_term_rtrancl_cstep_n_cases':
  assumes "linear_term s"
    and "(s \<cdot> \<sigma>, t) \<in> (cstep_n R n)\<^sup>*"
  shows "(\<exists>\<tau>. subst_domain \<tau> \<subseteq> vars_term s \<and>
    s \<cdot> \<tau> = t \<and> (\<forall>x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> (cstep_n R n)\<^sup>*)) \<or>
    (\<exists>\<tau> p u. subst_domain \<tau> \<subseteq> vars_term s \<and> (\<forall>x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> (cstep_n R n)\<^sup>*) \<and>
      p \<in> funposs s \<and> (s \<cdot> \<tau> |_ p, u) \<in> trs_n R n)"
using assms(2)
proof (induct)
  case base
  let ?\<tau> = "\<sigma> |s vars_term s"
  have "subst_domain ?\<tau> \<subseteq> vars_term s" by (simp add: subst_domain_vars_term_subset) 
  moreover have "s \<cdot> \<sigma> = s \<cdot> ?\<tau>" using coincidence_lemma by blast
  moreover have "\<forall>x \<in> vars_term s. (\<sigma> x, ?\<tau> x) \<in> (cstep_n R n)\<^sup>*" by auto
  ultimately show ?case by auto
next
  case (step t u)
  show ?case using step(3)
  proof
    assume "\<exists>\<tau>. subst_domain \<tau> \<subseteq> vars_term s \<and> s \<cdot> \<tau> = t \<and>
      (\<forall>x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> (cstep_n R n)\<^sup>*)"
    then obtain \<tau> where **: "subst_domain \<tau> \<subseteq> vars_term s"
      and t: "t = s \<cdot> \<tau>" and *: "\<forall>x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> (cstep_n R n)\<^sup>*" by blast
    then have "(s \<cdot> \<tau>, u) \<in> cstep_n R n" using \<open>(t, u) \<in> cstep_n R n\<close> by auto
    with \<open>linear_term s\<close>
      show ?thesis using * and **
    proof (cases rule: linear_term_cstep_n_cases)
      case (varposs \<mu>)
      then show ?thesis using * by (metis rtrancl_trans)
    qed blast
  qed blast
qed

lemma linear_term_rtrancl_cstep_n_cases [consumes 2]:
  assumes "linear_term s"
    and "(s \<cdot> \<sigma>, t) \<in> (cstep_n R n)\<^sup>*"
  obtains (varposs) \<tau> where "subst_domain \<tau> \<subseteq> vars_term s" and "s \<cdot> \<tau> = t"
    and "\<forall>x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> (cstep_n R n)\<^sup>*"
  | (funposs) \<tau> p u where "\<forall> x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> (cstep_n R n)\<^sup>*"
    and "p \<in> funposs s" and "(s \<cdot> \<tau> |_ p, u) \<in> trs_n R n"
using linear_term_rtrancl_cstep_n_cases' [OF assms] by blast

end
