(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Labelings_Impl
imports
  Labelings
  "../Rewriting/Trs_Impl"
begin

fun shows_lab' :: "('f \<Rightarrow> shows) \<Rightarrow> ('l \<Rightarrow> shows) \<Rightarrow> ('f, 'l) lab \<Rightarrow> shows" where
  "shows_lab' fun lab (UnLab f) = fun f"
| "shows_lab' fun lab (Lab f l) =
    (shows_lab' fun lab f +@+ ''['' +#+ lab l +@+ shows_string '']'')"
| "shows_lab' fun lab (Sharp f) = (shows_lab' fun lab f +@+ shows_string ''#'')"
| "shows_lab' fun lab (FunLab f l) =
    (shows_lab' fun lab f +@+ ''['' +#+ shows_list_gen (shows_lab' fun lab) [] [] [] [] l +@+ shows_string '']'')"

instantiation lab :: ("show", "show") "show"
begin

abbreviation "shows_lab \<equiv> shows_lab' shows shows"
definition "shows_prec_lab (d::nat) l = shows_lab l"

lemma shows_prec_lab_append [show_law_simps]:
  "shows_prec p (f :: ('a, 'b) lab) (r @ s) = shows_prec p f r @ s"
  by (induct f arbitrary: r s) (auto simp: shows_prec_lab_def show_law_simps)

definition "shows_list (fs :: ('a, 'b) lab list) = showsp_list shows_prec 0 fs"

instance by (default) (simp_all add: show_law_simps shows_list_lab_def)

end

text \<open>Resolve problems with recursive dependencies in generated code.\<close>
lemma [code]:
  "HOL.eq (FunLab f ls) (FunLab g ks) \<longleftrightarrow> HOL.eq f g \<and> list_all2 HOL.eq ls ks"
  by (simp_all add: list_all2_eq [symmetric])

abbreviation unlab_rules :: "(('f, 'l) lab, 'v) rules \<Rightarrow> (('f, 'l) lab, 'v) rules" where
  "unlab_rules \<equiv> map (map_funs_rule unlab)"

abbreviation init_lab_rules :: "('f, 'v) rules \<Rightarrow> (('f, 'l) lab, 'v) rules" where
  "init_lab_rules \<equiv> map init_lab_rule"

end

