(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)

section \<open>Level Confluence\<close>

theory Level_Confluence
imports
  "../Rewriting/Renaming_Interpretations"
  "../Conditional_Rewriting/Conditional_Rewriting"
  "../Rewriting/Multihole_Context"
begin

(*TODO: move*)
lemma csteps_n_mono:
  assumes "m \<le> n"
  shows "(cstep_n R m)\<^sup>* \<subseteq> (cstep_n R n)\<^sup>*"
using assms by (intro rtrancl_mono cstep_n_mono)

(*TODO: move*)
lemma greaterThanLessThan_all_conv:
  "(\<forall>i\<in>{m <..< n}. P i) \<longleftrightarrow> (\<forall>i. m < i \<and> i < n \<longrightarrow> P i)"
by auto

(*TODO: move*)
lemma in_NF_trs_permute [simp]:
  "\<pi> \<bullet> t \<in> NF_trs R \<longleftrightarrow> t \<in> NF_trs R"
by (metis NF_instance term_variants_iff)


subsection \<open>Preliminaries\<close>

(*TODO: move to AFP?*)
text \<open>Short notation for (diamond) commutation of two relations.\<close>
definition "comm r s \<longleftrightarrow>
  (\<forall>x y z. (x, y) \<in> r \<and> (x, z) \<in> s \<longrightarrow> (\<exists>u. (y, u) \<in> s \<and> (z, u) \<in> r))"

lemma commI [intro?]:
  assumes "\<And>x y z. \<lbrakk>(x, y) \<in> r; (x, z) \<in> s\<rbrakk> \<Longrightarrow> (\<exists>u. (y, u) \<in> s \<and> (z, u) \<in> r)"
  shows "comm r s"
  using assms by (auto simp: comm_def)

lemma commE:
  assumes "comm r s" and "(x, y) \<in> r" and "(x, z) \<in> s"
  obtains u where "(y, u) \<in> s" and "(z, u) \<in> r"
  using assms by (force simp: comm_def)

lemma comm_swap:
  "comm r s \<longleftrightarrow> comm s r"
  by (auto elim: commE intro!: commI)

lemma comm_rtrancl:
  assumes "comm r s"
  shows "comm (r\<^sup>*) s"
proof
  fix x y z
  assume "(x, y) \<in> r\<^sup>*" and "(x, z) \<in> s"
  then show "\<exists>u. (y, u) \<in> s \<and> (z, u) \<in> r\<^sup>*" (is "\<exists>u. ?P u y z")
  proof (induct)
    case base
    moreover have "(z, z) \<in> r\<^sup>*" by blast
    ultimately show ?case by blast
  next
    case (step a b)
    then obtain u where "(a, u) \<in> s" and "(z, u) \<in> r\<^sup>*" by blast
    moreover with \<open>(a, b) \<in> r\<close> and \<open>comm r s\<close> obtain v
      where "(b, v) \<in> s" and "(u, v) \<in> r\<^sup>*" by (blast elim: commE)
    ultimately have "(b, v) \<in> s" and "(z, v) \<in> r\<^sup>*" by auto
    then show ?case by blast
  qed
qed

interpretation crule_pt: prod_pt rule_pt.permute_prod rules_pt.permute_list ..

adhoc_overloading
  PERMUTE crule_pt.permute_prod and
  FRESH crule_pt.fresh crule_pt.fresh_set

interpretation ctrs_pt: set_pt crule_pt.permute_prod ..

adhoc_overloading
  PERMUTE ctrs_pt.permute_set and
  FRESH ctrs_pt.fresh

lemma finite_crule_supp:
  "finite (crule_pt.supp ((l, r), cs))"
  by (simp add: finite_term_supp rule_fs.finite_supp)

interpretation crule_fs: fs crule_pt.permute_prod
  by (default) (auto simp: finite_rule_supp finite_term_supp)

lemma supp_vars_crule_eq:
  fixes r :: "('f, 'v\<Colon>infinite) crule"
  shows "crule_pt.supp r = vars_crule r"
by (cases r; case_tac a)
   (auto simp: vars_crule_def vars_trs_def supp_vars_term_eq vars_rule_def)

lemma vars_crule_disjoint:
  fixes l r u v :: "('f, 'v\<Colon>infinite) term"
    and cs ds :: "('f, 'v) rule list"
  shows "\<exists>p. vars_crule (p \<bullet> ((l, r), cs)) \<inter> vars_crule ((u, v), ds) = {}"
proof -
  from crule_fs.supp_fresh_set obtain p
    where "crule_pt.supp (p \<bullet> ((l, r), cs)) \<sharp> ((u, v), ds)" by blast
  from crule_pt.fresh_set_disjoint [OF this]
    show ?thesis
    by (auto simp: supp_vars_crule_eq vars_rule_def)
qed

lemma X_vars_eqvt [eqvt]:
  "p \<bullet> X_vars r i = X_vars (p \<bullet> r) i"
proof -
  have "take i (snd (p \<bullet> r)) = p \<bullet> take i (snd r)"
    by (metis (no_types, hide_lams) crule_pt.snd_eqvt rules_pt.permute_list_def take_map)
  then show ?thesis
    by (simp_all add: X_vars_def eqvt)
qed

(*TODO: move*)
subsection \<open>Conditional Critical Pairs\<close>

definition "overlap R r r' p \<longleftrightarrow>
  (\<exists>p. p \<bullet> r \<in> R) \<and> (\<exists>p. p \<bullet> r' \<in> R) \<and> vars_crule r \<inter> vars_crule r' = {} \<and>
  p \<in> funposs (clhs r) \<and> (\<exists>\<sigma>. mgu (clhs r |_ p) (clhs r') = Some \<sigma>)"

definition
  CCP :: "('f, 'v\<Colon>infinite) ctrs \<Rightarrow> ('f, 'v) crule set"
where
  "CCP R = {((replace_at (clhs r') p (crhs r) \<cdot> \<sigma>, crhs r' \<cdot> \<sigma>),
    map (\<lambda>(s, t). (s \<cdot> \<sigma>, t \<cdot> \<sigma>)) (snd r @ snd r')) | r r' \<sigma> p.
  overlap R r r' p \<and> (p = \<epsilon> \<longrightarrow> \<not> (\<exists>p. p \<bullet> r = r'))}"

lemma overlapI [intro]:
  assumes "\<exists>p. p \<bullet> r \<in> R" and "\<exists>p. p \<bullet> r' \<in> R"
    and "vars_crule r \<inter> vars_crule r' = {}"
    and "p \<in> funposs (clhs r)"
    and "mgu (clhs r |_ p) (clhs r') = Some \<sigma>"
  shows "overlap R r r' p"
using assms unfolding overlap_def by blast

lemma CCP_I:
  assumes "\<exists>p. p \<bullet> r \<in> R" and "\<exists>p. p \<bullet> r' \<in> R"
    and "vars_crule r \<inter> vars_crule r' = {}"
    and "p \<in> funposs (clhs r)"
    and "mgu (clhs r |_ p) (clhs r') = Some \<sigma>"
    and "p = \<epsilon> \<Longrightarrow> \<not> (\<exists>p. p \<bullet> r = r')"
  shows "((replace_at (clhs r') p (crhs r) \<cdot> \<sigma>, crhs r' \<cdot> \<sigma>),
    map (\<lambda>(s, t). (s \<cdot> \<sigma>, t \<cdot> \<sigma>)) (snd r @ snd r')) \<in> CCP R"
using assms unfolding CCP_def by blast

definition right_stable :: "('f, 'v) ctrs \<Rightarrow> bool"
where
  "right_stable R \<longleftrightarrow> (\<forall> \<rho> \<in> R. \<forall> i < length (snd \<rho>).
    let t\<^sub>i = snd (snd \<rho> ! i) in
    (vars_term (clhs \<rho>) \<union> \<Union>(vars_term ` lhss (set (take (Suc i) (snd \<rho>)))) \<union>
    \<Union>(vars_term ` (rhss (set (take i (snd \<rho>)))))) \<inter> vars_term t\<^sub>i = {} \<and>
    (linear_term t\<^sub>i \<and> funas_term t\<^sub>i \<subseteq> funas_ctrs R - { f. defined (Ru R) f } \<or>
    ground t\<^sub>i \<and> t\<^sub>i \<in> NF (rstep (Ru R))))"

definition properly_oriented
where
  "properly_oriented R \<longleftrightarrow> (\<forall>\<rho> \<in> R.
    vars_term (crhs \<rho>) \<subseteq> vars_term (clhs \<rho>) \<or>
    (\<forall>i < length (snd \<rho>). vars_term (fst (snd \<rho> ! i)) \<subseteq> X_vars \<rho> i))"

(*
definition extended_properly_oriented
where
 "extended_properly_oriented R \<longleftrightarrow> (\<forall>\<rho> \<in> R. vars_term (crhs \<rho>) \<subseteq> vars_term (clhs \<rho>) \<or> (\<exists>m < length (snd \<rho>).
    (\<forall>i \<le> m. vars_term (fst (snd \<rho> ! i)) \<subseteq> X_vars \<rho> i) \<and>
    (\<forall>i \<in> {m <..< length (snd \<rho>)}.
      vars_term (crhs \<rho>) \<inter> (vars_term (fst (snd \<rho> ! i)) \<union> vars_term (snd (snd \<rho> ! i))) \<subseteq>
      X_vars \<rho> (Suc m))))"
*)

text \<open>
  Almost orthogonality modulo infeasibility (that is, all overlaps are either trivial,
  we have a root-overlap of two variants of the same rule,
  or the corresponding conditions are not all satisfiable at the same time).
\<close>
definition almost_orthogonal :: "('f, 'v\<Colon>infinite) ctrs \<Rightarrow> bool"
where
  "almost_orthogonal R \<longleftrightarrow>
    left_linear_trs (Ru R) \<and> (\<forall>r r' p. overlap R r r' p \<longrightarrow>
      p = \<epsilon> \<and> (\<exists>\<sigma>. mgu (clhs r) (clhs r') = Some \<sigma> \<and> crhs r \<cdot> \<sigma> = crhs r' \<cdot> \<sigma>) \<or>
      p = \<epsilon> \<and> (\<exists>p. p \<bullet> r = r') \<or>
      (\<forall>m n. comm ((cstep_n R m)\<^sup>*) ((cstep_n R n)\<^sup>*) \<longrightarrow>
        \<not> (\<exists>\<sigma>. conds_sat R m (snd r) \<sigma> \<and> conds_sat R n (snd r') \<sigma>)))"

lemma conds_sat_mono:
  "m \<le> n \<Longrightarrow> conds_sat R m cs \<sigma> \<Longrightarrow> conds_sat R n cs \<sigma>"
using csteps_n_mono [of m n R]
by (auto simp: conds_sat_iff)

text \<open>The usual definition of infeasibility is a sufficient condition.\<close>
lemma
  assumes "\<not> (\<exists>\<sigma> n. conds_sat R n (snd r @ snd r') \<sigma>)"
  shows "\<forall>m n. comm ((cstep_n R m)\<^sup>*) ((cstep_n R n)\<^sup>*) \<longrightarrow>
        \<not> (\<exists>\<sigma>. conds_sat R m (snd r) \<sigma> \<and> conds_sat R n (snd r') \<sigma>)"
proof -
  { fix m n
    have "\<not> (\<exists>\<sigma>. conds_sat R m (snd r) \<sigma> \<and> conds_sat R n (snd r') \<sigma>)"
    using assms by (auto dest!: conds_sat_mono [of m "max m n", simplified] conds_sat_mono [of n "max m n", simplified]) }
  then show ?thesis by blast
qed

definition level_confluent :: "('f, 'v) ctrs \<Rightarrow> bool"
where
  "level_confluent R \<longleftrightarrow> (\<forall> n. CR (cstep_n R n))"

lemma almost_orthogonal_imp_linear [simp]:
  "almost_orthogonal R \<Longrightarrow> ((l, r), cs) \<in> R \<Longrightarrow> linear_term l"
  by (auto simp: almost_orthogonal_def left_linear_trs_def Ru_def)

lemma almost_orthogonal_overlap_cases [consumes 2, cases pred: almost_orthogonal]:
  assumes "almost_orthogonal R" and "overlap R r r' p"
  obtains (trivial) \<sigma> where "p = \<epsilon>" and "mgu (clhs r) (clhs r') = Some \<sigma>"
    and "crhs r \<cdot> \<sigma> = crhs r' \<cdot> \<sigma>"
  | (variants) \<pi> where "p = \<epsilon>" and "\<pi> \<bullet> r = r'"
  | (infeasible) "\<And>m n. comm ((cstep_n R m)\<^sup>*) ((cstep_n R n)\<^sup>*) \<Longrightarrow>
      \<not> (\<exists>\<sigma>. conds_sat R m (snd r) \<sigma> \<and> conds_sat R n (snd r') \<sigma>)"
using assms unfolding almost_orthogonal_def by blast

(*unused*)
lemma cstep_n_varpeak:
  assumes "CCP R = {}" and rule1: "((l, r), cs) \<in> R"
    and "linear_term l"
    and l_\<sigma>: "l \<cdot> \<sigma> = C\<langle>t\<rangle>" and "C \<noteq> \<box>" and cstep: "(t, u) \<in> cstep_n R n"
  shows "\<exists>\<tau>. l \<cdot> \<tau> = C\<langle>u\<rangle> \<and> subst_domain \<tau> \<subseteq> vars_term l \<and>
    (\<forall>x \<in> vars_term l. (\<sigma> x, \<tau> x) \<in> (cstep_n R n)\<^sup>*)"
proof -
  from cstep [THEN cstep_nE] obtain D l\<^sub>2 r\<^sub>2 \<sigma>\<^sub>2 cs\<^sub>2 k
    where rule2: "((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R" and [simp]: "n = Suc k"
    and conds: "conds_sat R k cs\<^sub>2 \<sigma>\<^sub>2"
    and t: "t = D\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>" and u: "u = D\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>"
    unfolding conds_sat_iff by metis (*only metis works; investigate!*)
  def p \<equiv> "hole_pos C <#> hole_pos D"
  have *: "p \<in> poss (l \<cdot> \<sigma>)" by (simp add: assms t p_def)
  show ?thesis
  proof (cases "p \<in> funposs l")
    case False
    with * obtain q\<^sub>1 q\<^sub>2 x
      where p: "p = q\<^sub>1 <#> q\<^sub>2" and q\<^sub>1: "q\<^sub>1 \<in> poss l"
      and lq\<^sub>1: "l |_ q\<^sub>1 = Var x" and q\<^sub>2: "q\<^sub>2 \<in> poss (\<sigma> x)"
        by (rule poss_subst_apply_term)
    moreover
    have [simp]: "l \<cdot> \<sigma> |_ p = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using assms p_def t by auto 
    ultimately
    have [simp]: "\<sigma> x |_ q\<^sub>2 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" by simp
    
    def \<tau> \<equiv> "\<lambda>y. if y = x then replace_at (\<sigma> x) q\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2) else (\<sigma> |s vars_term l) y"
    have \<tau>_x: "\<tau> x = replace_at (\<sigma> x) q\<^sub>2 (r\<^sub>2 \<cdot> \<sigma>\<^sub>2)" by (simp add: \<tau>_def)
    have [simp]: "\<And>y. y \<in> vars_term l \<Longrightarrow> y \<noteq> x \<Longrightarrow> \<tau> y = \<sigma> y" by (simp add: \<tau>_def)
    have "(\<sigma> x, \<tau> x) \<in> cstep_n R n"
    proof -
      let ?C = "ctxt_of_pos_term q\<^sub>2 (\<sigma> x)"
      have "(?C\<langle>l\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>, ?C\<langle>r\<^sub>2 \<cdot> \<sigma>\<^sub>2\<rangle>) \<in> cstep_n R n"
        using conds by (auto intro!: cstep_n_SucI rule2 simp: conds_sat_iff)
      then show ?thesis using q\<^sub>2 by (simp add: \<tau>_def replace_at_ident)
    qed
    then have "\<forall>x \<in> vars_term l. (\<sigma> x, \<tau> x) \<in> (cstep_n R n)\<^sup>*" by (auto simp: \<tau>_def)
    moreover have "subst_domain \<tau> \<subseteq> vars_term l"
      using lq\<^sub>1 and vars_term_subt_at [OF q\<^sub>1]
      by (auto simp: subst_domain_def \<tau>_def subst_restrict_def)
    moreover have "l \<cdot> \<tau> = C\<langle>u\<rangle>"
    proof -
      have [simp]: "ctxt_of_pos_term p (l \<cdot> \<sigma>) = C \<circ>\<^sub>c D"
        by (simp add: assms t p_def ctxt_of_pos_term_append)
      have "l \<cdot> \<sigma> |_ q\<^sub>1 = \<sigma> x" by (simp add: lq\<^sub>1 q\<^sub>1)
      from linear_term_replace_in_subst [OF \<open>linear_term l\<close> q\<^sub>1 lq\<^sub>1, of \<sigma> \<tau>, OF _ \<tau>_x, folded ctxt_compose_apply_distrib]
        and q\<^sub>1 and ctxt_of_pos_term_append [of q\<^sub>1 "l \<cdot> \<sigma>" q\<^sub>2, folded p, unfolded this, symmetric]
        show "l \<cdot> \<tau> = C\<langle>u\<rangle>" by (simp add: u)
    qed
    ultimately show ?thesis by blast
  next
    case True
    from vars_crule_disjoint obtain \<pi>\<^sub>1
      where \<pi>\<^sub>1: "vars_crule (\<pi>\<^sub>1 \<bullet> ((l, r), cs)) \<inter> vars_crule ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) = {}" ..
    def l\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> l" and r\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> r"
      and cs\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> cs" and \<sigma>\<^sub>1 \<equiv> "sop (-\<pi>\<^sub>1) \<circ>\<^sub>s \<sigma>"
    note rename = l\<^sub>1_def r\<^sub>1_def cs\<^sub>1_def \<sigma>\<^sub>1_def
  
    have *: "-\<pi>\<^sub>1 \<bullet> ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<in> R" and "0 \<bullet> ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R"
      using rule1 and rule2 by (simp_all add: eqvt rename o_def)
    then have rule_variants: "\<exists>\<pi>. \<pi> \<bullet> ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<in> R" "\<exists>\<pi>. \<pi> \<bullet> ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R" by blast+
  
    have disj: "vars_crule ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<inter> vars_crule ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) = {}"
      using \<pi>\<^sub>1 by (auto simp: eqvt rename)
  
    from True have funposs: "p \<in> funposs (clhs ((l\<^sub>1, r\<^sub>1), cs\<^sub>1))" by (simp add: rename)
    then have "p \<in> poss l\<^sub>1" by (auto dest: funposs_imp_poss)
    with vars_term_subt_at [OF this] and disj and l_\<sigma>
      have "vars_term (l\<^sub>1 |_ p) \<inter> vars_term l\<^sub>2 = {}" and "(l\<^sub>1 |_ p) \<cdot> \<sigma>\<^sub>1 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2"
      by (auto simp: eqvt vars_crule_def vars_rule_def rename)
         (metis hole_pos_poss p_def subt_at_append subt_at_hole_pos subt_at_subst t)
    from vars_term_disjoint_imp_unifier [OF this]
      obtain \<mu> where mgu: "mgu (l\<^sub>1 |_ p) l\<^sub>2 = Some \<mu>"
      using mgu_complete by (auto simp: unifiers_def)
     
     have "p \<noteq> \<epsilon>" using \<open>C \<noteq> \<box>\<close> by (cases C) (simp_all add: p_def)
     with CCP_I [OF rule_variants disj funposs] and mgu and \<open>CCP R = {}\<close>
       show ?thesis by auto
  qed
qed


subsection \<open>Extended parallel rewriting\<close>

definition "epar_n_mctxt R n C =
  {(s, t). \<exists>ss ts. s =\<^sub>f (C, ss) \<and> t =\<^sub>f (C, ts) \<and>
    (\<forall>i<length ss. (ss ! i, ts ! i) \<in> trs_n R n \<union> (cstep_n R (n - 1))\<^sup>*)}"

definition "epar_n R n =
  {(s, t). \<exists>C ss ts. s =\<^sub>f (C, ss) \<and> t =\<^sub>f (C, ts) \<and>
    (\<forall>i<length ss. (ss ! i, ts ! i) \<in> trs_n R n \<union> (cstep_n R (n - 1))\<^sup>*)}"

lemma epar_n_epar_n_mctxt_conv:
  "(s, t) \<in> epar_n R n \<longleftrightarrow> (\<exists>C. (s, t) \<in> epar_n_mctxt R n C)"
  by (auto simp: epar_n_def epar_n_mctxt_def)

lemma epar_n_0 [simp]:
  "epar_n R 0 = Id"
by (auto simp: epar_n_def eq_fill.simps list_eq_iff_nth_eq intro: arg_cong)
   (metis eqfE mctxt_of_term)

lemma epar_n_SucE:
  assumes "(s, t) \<in> epar_n R (Suc n)"
  obtains C ss ts where "s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)"
    and "length ss = length ts"
    and "\<forall>i<length ss. (ss ! i, ts ! i) \<in> trs_n R (Suc n) \<union> (cstep_n R n)\<^sup>*"
proof -
  obtain C ss ts where "s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)"
    and "\<forall>i<length ss. (ss ! i, ts ! i) \<in> trs_n R (Suc n) \<union> (cstep_n R n)\<^sup>*"
    using assms by (auto simp: epar_n_def)
  then have "length ss = length ts" by (auto dest!: eqfE)
  show ?thesis by (rule) fact+
qed

lemma epar_n_mctxt_SucE:
  assumes "(s, t) \<in> epar_n_mctxt R (Suc n) C"
  obtains ss ts where "s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)"
    and "length ss = length ts"
    and "\<forall>i<length ss. (ss ! i, ts ! i) \<in> trs_n R (Suc n) \<union> (cstep_n R n)\<^sup>*"
using assms by (auto simp: epar_n_mctxt_def elim!: eq_fill.cases) (metis eqf_refl)

lemma epar_n_mctxtI:
  assumes "s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)"
    and "\<And>i. i < length ss \<Longrightarrow> (ss ! i, ts ! i) \<in> trs_n R n \<union> (cstep_n R (n - 1))\<^sup>*"
  shows "(s, t) \<in> epar_n_mctxt R n C"
  using assms by (auto simp: epar_n_mctxt_def)

lemma epar_nI:
  assumes "s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)"
    and "\<And>i. i < length ss \<Longrightarrow> (ss ! i, ts ! i) \<in> trs_n R n \<union> (cstep_n R (n - 1))\<^sup>*"
  shows "(s, t) \<in> epar_n R n"
  using assms by (auto simp: epar_n_def)

lemma trs_n_subset_epar_n:
  "trs_n R n \<subseteq> epar_n R n"
  by (auto intro!: epar_nI [of s MHole "[s]" t "[t]" R n for s t])

lemma epar_n_par_rstep_1_conv:
  "epar_n R 1 = par_rstep (trs_n R 1)"
  by (auto simp: epar_n_def) (force intro: par_rstep_mctxt par_rstep_id dest: par_rstep_mctxtD)+

lemma cstep_n_subset_epar_n:
  "cstep_n R n \<subseteq> epar_n R n"
proof
  fix s t
  assume "(s, t) \<in> cstep_n R n"
  then obtain C \<sigma> l r where [simp]: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and [simp]:"t = C\<langle>r \<cdot> \<sigma>\<rangle>"
    and "(l, r) \<in> trs_n R n"
    unfolding "cstep_n_rstep_trs_n_conv" by fast
  then have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> trs_n R n" using trs_n_subst by fast
  then show "(s, t) \<in> epar_n R n"
    by (auto intro: epar_nI[OF mctxt_of_ctxt mctxt_of_ctxt])
qed

lemma epar_n_NF_imp_cstep_n_NF: "t \<in> NF (epar_n R n) \<Longrightarrow> t \<in> NF (cstep_n R n)"
using assms NF_anti_mono [OF cstep_n_subset_epar_n] by blast

lemma epar_n_subset_csteps_n:
  "epar_n R n \<subseteq> (cstep_n R n)\<^sup>*"
proof (cases n)
  case (Suc k)
  then show ?thesis
    using rtrancl_mono [OF rstep_mono [OF trs_n_Suc_mono [of R k]]]
    by (auto elim!: epar_n_SucE simp: cstep_n_rstep_trs_n_conv) (rule rsteps_mctxt, auto+)
qed simp

lemma epars_n_subset_csteps_n: "(epar_n R n)\<^sup>* \<subseteq> (cstep_n R n)\<^sup>*"
using rtrancl_mono [OF epar_n_subset_csteps_n] rtrancl_idemp by auto

lemma epar_n_refl [simp]:
  "(t, t) \<in> epar_n R n"
  by (auto simp: epar_n_def intro: mctxt_of_term)

lemma csteps_n_subset_epar_n_Suc:
  "(cstep_n R n)\<^sup>* \<subseteq> epar_n R (Suc n)"
proof
  fix s t
  assume "(s, t) \<in> (cstep_n R n)\<^sup>*"
  then show "(s, t) \<in> epar_n R (Suc n)"
  proof (induct)
    case (step t u)
    then show ?case
      by (intro epar_nI [of _ MHole "[s]" _ "[u]"])
         (auto elim!: epar_n_SucE dest: rtrancl_into_rtrancl)
  qed simp
qed

lemma epar_n_Suc_mono:
  "epar_n R n \<subseteq> epar_n R (Suc n)"
using epar_n_subset_csteps_n csteps_n_subset_epar_n_Suc by blast

lemma rtrancl_epar_n_conv:
  "(epar_n R n)\<^sup>* = (cstep_n R n)\<^sup>*"
by (intro rtrancl_subset cstep_n_subset_epar_n epar_n_subset_csteps_n)

lemma epar_n_rtrancl_Suc_mono:
  "(epar_n R n)\<^sup>* \<subseteq> epar_n R (Suc n)"
using rtrancl_mono [OF epar_n_subset_csteps_n]
  and csteps_n_subset_epar_n_Suc by (simp) blast

lemma epar_n_mono:
  assumes "n \<le> m"
  shows "epar_n R n \<subseteq> epar_n R m"
using assms
proof (induct "m - n" arbitrary: n)
  case (Suc k)
  then have "k = m - (Suc n)" and "Suc n \<le> m" by arith+
  from Suc.hyps(1) [OF this]
    show ?case using epar_n_Suc_mono by blast
qed simp

definition "epar R = (\<Union>n. epar_n R n)"

lemma epar_iff:
  "(s, t) \<in> epar R \<longleftrightarrow> (\<exists>n. (s, t) \<in> epar_n R n)"
  by (auto simp: epar_def)

lemma all_ctxt_closed_epar_n [simp]:
  "all_ctxt_closed UNIV (epar_n R n)"
proof -
  { fix f ts ss
    assume [simp]: "length ts = length ss"
      and "\<forall>i<length ss. (ts ! i, ss ! i) \<in> epar_n R n"
    then have "\<forall>i<length ss. \<exists>C us vs.
      ts ! i =\<^sub>f (C, us) \<and> ss ! i =\<^sub>f (C, vs) \<and>
      (\<forall>i < length us. (us ! i, vs ! i) \<in> trs_n R n \<union> (cstep_n R (n - 1))\<^sup>*)"
      (is "\<forall>i<length ss. \<exists>C us vs. ?P i C us vs") by (auto simp: epar_n_def)
    then have "\<forall>i<length ss. \<exists>x. ?P i (fst x) (fst (snd x)) (snd (snd x))" by force
    then obtain c where "\<forall>i<length ss. ?P i (fst (c i)) (fst (snd (c i))) (snd (snd (c i)))"
      unfolding choice_iff' by blast
    moreover
    def Cs \<equiv> "map (fst \<circ> c) [0 ..< length ss]"
      and us \<equiv> "map (fst \<circ> snd \<circ> c) [0 ..< length ss]"
      and  vs \<equiv> "map (snd \<circ> snd \<circ> c) [0 ..< length ss]"
    ultimately have [simp]: "length Cs = length ts" "length us = length ts" "length vs = length ts"
      and *: "\<forall>i<length ss. ?P i (Cs ! i) (us ! i) (vs ! i)"
      by (simp_all)
    def C \<equiv> "MFun f Cs"
    have "Fun f ts =\<^sub>f (C, concat us)" and "Fun f ss =\<^sub>f (C, concat vs)"
    using * by (auto simp: C_def intro: eqf_MFunI)
  moreover have "\<forall>i<length (concat us). (concat us ! i, concat vs ! i) \<in> trs_n R n \<union> (cstep_n R (n - 1))\<^sup>*"
    using * by (intro concat_all_nth) (auto simp: eq_fill.simps)
  ultimately have "(Fun f ts, Fun f ss) \<in> epar_n R n" by (auto intro!: epar_nI) }
  then show ?thesis by (auto simp: all_ctxt_closed_def)
qed

lemma epar_n_mctxt:
  assumes "s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)"
    and "\<forall>i<length ss. (ss ! i, ts ! i) \<in> epar_n R n"
  shows "(s, t) \<in> epar_n R n"
proof -
  have [simp]: "length ss = length ts" using assms by (auto dest!: eqfE)
  have [simp]: "t = fill_holes C ts" using assms by (auto dest: eqfE)
  have "(s, fill_holes C ts) \<in> epar_n R n"
    using assms by (intro eqf_all_ctxt_closed_step [of UNIV _ s C ss, THEN conjunct1]) auto
  then show ?thesis by simp
qed

lemma epar_n_mctxt_mctxt:
  assumes "s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)" and "length Cs = length ss"
    and "\<forall>i < length ss. (ss ! i, ts ! i) \<in> epar_n_mctxt R n (Cs ! i)"
  shows "(s, t) \<in> epar_n_mctxt R n (fill_holes_mctxt C Cs)"
proof -
  have "\<forall>i < length ss. \<exists>us vs. ss ! i =\<^sub>f (Cs ! i, us) \<and> ts ! i =\<^sub>f (Cs ! i, vs) \<and>
    (\<forall>j < length us. (us ! j, vs ! j) \<in> trs_n R n \<union> (cstep_n R (n - 1))\<^sup>*)"
    using assms by (auto simp: epar_n_mctxt_def)
  then obtain f and g where *: "\<forall>i < length ss. ss ! i =\<^sub>f (Cs ! i, f i) \<and> ts ! i =\<^sub>f (Cs ! i, g i)"
    and **: "\<forall>i < length ss.
    (\<forall>j < length (f i). (f i ! j, g i ! j) \<in> trs_n R n \<union> (cstep_n R (n - 1))\<^sup>*)" by metis
  moreover def sss \<equiv> "map f [0 ..< length ss]" and tss \<equiv> "map g [0 ..< length ss]"
  ultimately have "s =\<^sub>f (fill_holes_mctxt C Cs, concat sss)"
    and "t =\<^sub>f (fill_holes_mctxt C Cs, concat tss)"
    using assms by (auto intro: fill_holes_mctxt_sound elim!: eq_fill.cases)
  moreover
  { fix i
    assume "i < length (concat sss)"
    from less_length_concat [OF this] obtain j k
      where "j < length sss" and "k < length (sss ! j)"
      and "i = listsum (map length (take j sss)) + k"
      and "concat sss ! i = sss ! j ! k" by blast
    moreover then have "j < length tss" and "k < length (tss ! j)"
      and "i = listsum (map length (take j tss)) + k"
      using *
      by (auto simp: sss_def tss_def dest!: eqfE intro!: arg_cong [of _ _ listsum] nth_equalityI)
    moreover from concat_nth [OF this] have "concat tss ! i = tss ! j ! k" .
    ultimately have "(concat sss ! i, concat tss ! i) \<in> trs_n R n \<union> (cstep_n R (n - 1))\<^sup>*"
      using ** by (auto simp: sss_def tss_def) }
  ultimately show ?thesis by (rule epar_n_mctxtI)
qed

lemma Var_epar_n [dest]:
  assumes "(Var x, t) \<in> epar_n R n"
    and "\<forall>((l, r), cs) \<in> R. is_Fun l"
  shows "t = Var x"
using assms
by (cases n)
   (force elim!: epar_n_SucE eq_fill.cases fill_holes_eq_Var_cases intro: fill_holes_MHole)+

lemma subst_epar_n_imp_epar_n:
  fixes \<sigma> \<tau> :: "('f, 'v) subst"
  assumes "\<forall>x \<in> vars_term t. (\<sigma> x, \<tau> x) \<in> epar_n R n"
  shows "(t \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> epar_n R n"
using assms by (intro all_ctxt_closed_subst_step) simp_all

lemma epar_n_ctxt:
  assumes "(s, t) \<in> epar_n R n"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> epar_n R n"
using epar_n_mctxt [of "C\<langle>s\<rangle>" "mctxt_of_ctxt C" "[s]" "C\<langle>t\<rangle>" "[t]" R n]
  and assms by simp

lemma epar_n_rtrancl_all_ctxt_closed [simp]:
  "all_ctxt_closed UNIV ((epar_n R n)\<^sup>*)"
by (rule trans_ctxt_imp_all_ctxt_closed)
   (auto simp: ctxt_closed_def elim!: ctxt_closure.cases intro: epar_n_ctxt [THEN rtrancl_map])

lemma subst_epar_n_imp_epar_n_rtrancl:
  fixes \<sigma> \<tau> :: "('f, 'v) subst"
  assumes "\<forall>x \<in> vars_term t. (\<sigma> x, \<tau> x) \<in> (epar_n R n)\<^sup>*"
  shows "(t \<cdot> \<sigma>, t \<cdot> \<tau>) \<in> (epar_n R n)\<^sup>*"
using assms by (intro all_ctxt_closed_subst_step) simp_all

lemma linear_term_epar_n_mctxt_cases':
  assumes "linear_term s"
  and "(s \<cdot> \<sigma>, t) \<in> epar_n_mctxt R (Suc n) C"
  shows "(\<exists>\<tau>. subst_domain \<tau> \<subseteq> vars_term s \<and> t = s \<cdot> \<tau> \<and>
      (\<forall>x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> epar_n R (Suc n))) \<or>
    (\<exists>\<tau> u p q. p \<in> hole_poss C \<and> p <#> q \<in> funposs s \<and> (s \<cdot> \<tau> |_ (p <#> q), u) \<in> trs_n R (Suc n))"
    (is "(\<exists>\<tau>. ?P1 \<tau> s t) \<or> _" is "?P C s t")
using assms
proof (induction s arbitrary: C t)
  case (Var x t)
  then have "?P1 (subst x t) (Var x) t" by (auto simp: epar_n_epar_n_mctxt_conv)
  then show ?case by blast
next
  case (Fun f ss)
  from epar_n_mctxt_SucE [OF Fun.prems(2)] obtain ts us
    where eqf: "Fun f ss \<cdot> \<sigma> =\<^sub>f (C, ts)" "t =\<^sub>f (C, us)" and [simp]: "length ts = length us"
    and *: "\<forall>i < length ts. (ts ! i, us ! i) \<in> trs_n R (Suc n) \<union> (cstep_n R n)\<^sup>*" by blast
  show ?case
  proof (cases "C = MHole")
    case True
    with * have "(Fun f ss \<cdot> \<sigma>, t) \<in> trs_n R (Suc n) \<union> (cstep_n R n)\<^sup>*"
      using eqf by (cases ts) (auto elim!: eq_fill.cases simp: Suc_length_conv)
    then show ?thesis
    proof
      assume "(Fun f ss \<cdot> \<sigma>, t) \<in> trs_n R (Suc n)"
      with True have "\<epsilon> \<in> hole_poss C"
        and "\<epsilon> <#> \<epsilon> \<in> funposs (Fun f ss)" and "(Fun f ss \<cdot> \<sigma> |_ (\<epsilon> <#> \<epsilon>), t) \<in> trs_n R (Suc n)" by auto
      then show ?thesis by blast
    next
      assume "(Fun f ss \<cdot> \<sigma>, t) \<in> (cstep_n R n)\<^sup>*"
      with Fun.prems(1) show ?thesis
      proof (cases rule: linear_term_rtrancl_cstep_n_cases)
        case (varposs \<tau>)
        then show ?thesis using csteps_n_subset_epar_n_Suc [of R n] by blast
      next
        case (funposs \<tau> q u)
        moreover then have "\<epsilon> <#> q \<in> funposs (Fun f ss)"
          and "(Fun f ss \<cdot> \<tau> |_ (\<epsilon> <#> q), u) \<in> trs_n R n" by simp_all
        moreover have "\<epsilon> \<in> hole_poss C" by (simp add: True)
        ultimately show ?thesis using trs_n_Suc_mono [of R n] by blast
      qed
    qed
  next
    case False
    then obtain Cs where C: "C = MFun f Cs"
      and [simp]: "length Cs = length ss" "listsum (map num_holes Cs) = length ts"
      and eqfs: "\<forall>i < length ss. (ss ! i) \<cdot> \<sigma> =\<^sub>f (Cs ! i, partition_holes ts Cs ! i)"
      using eqf by (cases C) (auto dest!: eqf_Fun_MFun elim: eq_fill.cases)
    let ?ts = "partition_holes ts Cs"
    let ?us = "partition_holes us Cs"
    let ?u = "\<lambda>i. fill_holes (Cs ! i) (?us ! i)"
    have "\<forall>i < length ts. (ts ! i, us ! i) \<in> epar_n_mctxt R (Suc n) MHole"
      using * by (auto intro!: epar_n_mctxtI [of s MHole "[s]" t "[t]" for s t])
    then have "\<forall>i < length ts. (concat ?ts ! i, concat ?us ! i) \<in> epar_n_mctxt R (Suc n) MHole"
      by (simp)
    then have **: "\<forall>i < length ss. (\<forall>j < length (?ts ! i). (?ts ! i ! j, ?us ! i ! j) \<in> epar_n_mctxt R (Suc n) MHole)"
      using concat_nth_length [of _ ?ts]
      by (auto simp: concat_nth_length partition_by_nth_nth take_map [symmetric])
    { fix i assume "i < length ss"
      then have "((ss ! i) \<cdot> \<sigma>, ?u i) \<in>
        epar_n_mctxt R (Suc n) (fill_holes_mctxt (Cs ! i) (replicate (num_holes (Cs ! i)) MHole))"
        using eqfs and ** by (intro epar_n_mctxt_mctxt) (auto elim!: eq_fill.cases) }
    with Fun(1, 2) have IH: "\<forall>i < length ss. ?P (Cs ! i) (ss ! i) (?u i)" by auto
    show ?thesis
    proof (cases "\<forall>i < length ss. (\<exists>\<tau>. ?P1 \<tau> (ss ! i) (?u i))")
      case False
      then obtain i and \<tau> and u and p and q where "i < length ss"
        and "p \<in> hole_poss (Cs ! i)" and "p <#> q \<in> funposs (ss ! i)"
        and "((ss ! i) \<cdot> \<tau> |_ (p <#> q), u) \<in> trs_n R (Suc n)"
        using IH by blast
      moreover then have "i <# p \<in> hole_poss C" and "(i <# p) <#> q \<in> funposs (Fun f ss)"
        and "(Fun f ss \<cdot> \<tau> |_ ((i <# p) <#> q), u) \<in> trs_n R (Suc n)" by (auto simp: C)
      ultimately show ?thesis by blast
    next
      case True
      then obtain \<tau> where **: "\<forall>i < length ss. ?P1 (\<tau> i) (ss ! i) (?u i)"
        unfolding choice_iff' [of "\<lambda>i. i < length ss"] by auto
      have "is_partition (map vars_term ss)" using \<open>linear_term (Fun f ss)\<close> by simp
      from subst_merge [OF this, of \<tau>] obtain \<mu>
        where [simp]: "\<And>i x. i < length ss \<Longrightarrow> x \<in> vars_term (ss ! i) \<Longrightarrow> \<mu> x = \<tau> i x" by blast
      then have [simp]: "\<And>i. i < length ss \<Longrightarrow> ss ! i \<cdot> \<tau> i = ss ! i \<cdot> \<mu>"
        by (simp add: term_subst_eq_conv)
      let ?\<mu> = "\<mu> |s (vars_term (Fun f ss))"
      have "subst_domain ?\<mu> \<subseteq> vars_term (Fun f ss)" by (metis subst_domain_vars_term_subset)
      moreover have "t = Fun f ss \<cdot> ?\<mu>"
      proof -
        from eqf have "t = fill_holes C us" by (auto dest: eqfE)
        moreover have "\<forall>i < length Cs. fill_holes (Cs ! i) (?us ! i) = (ss ! i) \<cdot> ?\<mu>"
          using ** by (auto intro!: term_subst_eq simp: subst_restrict_def)
        ultimately show ?thesis
          by (auto simp del: fill_holes.simps simp: C partition_holes_fill_holes_conv
                   intro: nth_map_conv)
      qed
      moreover have "\<forall>x \<in> vars_term (Fun f ss). (\<sigma> x, ?\<mu> x) \<in> epar_n R (Suc n)"
        using ** by (auto dest: in_set_idx)
      ultimately show ?thesis by blast
    qed
  qed
qed

lemma linear_term_epar_n_mctxt_cases [consumes 2]:
  assumes "linear_term s"
  and "(s \<cdot> \<sigma>, t) \<in> epar_n_mctxt R (Suc n) C"
  obtains (varposs) \<tau> where "subst_domain \<tau> \<subseteq> vars_term s" and "t = s \<cdot> \<tau>"
    and "\<forall>x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> epar_n R (Suc n)"
  | (funposs) \<tau> u p q where "p \<in> hole_poss C" and "p <#> q \<in> funposs s"
    and "(s \<cdot> \<tau> |_ (p <#> q), u) \<in> trs_n R (Suc n)"
using linear_term_epar_n_mctxt_cases' [OF assms] by blast

lemma linear_term_epar_n_cases [consumes 2]:
  assumes "linear_term s"
  and "(s \<cdot> \<sigma>, t) \<in> epar_n R (Suc n)"
  obtains (varposs) \<tau> where "subst_domain \<tau> \<subseteq> vars_term s" and "t = s \<cdot> \<tau>"
    and "\<forall>x \<in> vars_term s. (\<sigma> x, \<tau> x) \<in> epar_n R (Suc n)"
  | (funposs) \<tau> u p where "p \<in> funposs s"
    and "(s \<cdot> \<tau> |_ p, u) \<in> trs_n R (Suc n)"
using assms by (auto simp: epar_n_epar_n_mctxt_conv elim!: linear_term_epar_n_mctxt_cases)

lemma epar_n_varpeak':
  assumes comm: "comm ((cstep_n R m)\<^sup>*) ((cstep_n R n)\<^sup>*)"
  assumes ao: "almost_orthogonal R" and rule: "((l, r), cs) \<in> R"
    and conds: "conds_sat R m cs \<sigma>"
    and l_\<sigma>: "l \<cdot> \<sigma> =\<^sub>f (C, ts)" and "C \<noteq> MHole"
    and epar: "\<forall>i<length ts. (ts ! i, us ! i) \<in> trs_n R (Suc n) \<union> (cstep_n R n)\<^sup>*"
    and len: "length us = length ts"
  shows "\<exists>\<tau>. l \<cdot> \<tau> =\<^sub>f (C, us) \<and> subst_domain \<tau> \<subseteq> vars_term l \<and>
    (\<forall>x \<in> vars_term l. (\<sigma> x, \<tau> x) \<in> epar_n R (Suc n))"
proof -
  have "linear_term l" using assms by simp
  have "(l \<cdot> \<sigma>, fill_holes C us) \<in> epar_n_mctxt R (Suc n) C"
    using epar and l_\<sigma> and len
    by (intro epar_n_mctxtI [OF l_\<sigma>, of "fill_holes C us" us]) (auto elim!: eq_fill.cases)
  with \<open>linear_term l\<close> show ?thesis
  proof (cases rule: linear_term_epar_n_mctxt_cases)
    case (varposs \<tau>)
    moreover then have "l \<cdot> \<tau> =\<^sub>f (C, us)" using l_\<sigma> and len by (auto elim!: eq_fill.cases)
    ultimately show ?thesis by blast
  next
    case (funposs \<tau> u p q)
    obtain l\<^sub>2 r\<^sub>2 cs\<^sub>2 \<sigma>\<^sub>2
      where l_\<tau>: "l \<cdot> \<tau> |_ (p <#> q) = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" and u: "u = r\<^sub>2 \<cdot> \<sigma>\<^sub>2"
      and rule2: "((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R"
      and conds2: "conds_sat R n cs\<^sub>2 \<sigma>\<^sub>2"
        by (rule funposs(3) [THEN trs_n_SucE]) (auto simp: cstep_n_rstep_trs_n_conv conds_sat_iff)

    from vars_crule_disjoint obtain \<pi>\<^sub>1
      where \<pi>\<^sub>1: "vars_crule (\<pi>\<^sub>1 \<bullet> ((l, r), cs)) \<inter> vars_crule ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) = {}" ..
    def l\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> l" and r\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> r"
      and cs\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> cs" and \<sigma>\<^sub>1 \<equiv> "sop (-\<pi>\<^sub>1) \<circ>\<^sub>s \<tau>"
    note rename = l\<^sub>1_def r\<^sub>1_def cs\<^sub>1_def \<sigma>\<^sub>1_def

    have *: "-\<pi>\<^sub>1 \<bullet> ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<in> R" and "0 \<bullet> ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R"
      using rule and rule2 by (simp_all add: eqvt rename o_def)
    then have rule_variants: "\<exists>\<pi>. \<pi> \<bullet> ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<in> R" "\<exists>\<pi>. \<pi> \<bullet> ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R" by blast+
  
    have disj: "vars_crule ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<inter> vars_crule ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) = {}"
      using \<pi>\<^sub>1 by (auto simp: eqvt rename)

    def \<delta> \<equiv> "\<lambda>x. if x \<in> \<Union>(vars_rule ` set cs\<^sub>1) then \<sigma> (-\<pi>\<^sub>1 \<bullet> x) else \<sigma>\<^sub>2 x"

    have *: "\<Union>(vars_rule ` set cs\<^sub>1) \<inter> \<Union>(vars_rule ` set cs\<^sub>2) = {}"
      using disj by (auto simp: vars_crule_def vars_trs_def)

    { fix s t assume "(s, t) \<in> set cs\<^sub>1"
      then have "s \<cdot> \<delta> = s \<cdot> sop (-\<pi>\<^sub>1) \<circ>\<^sub>s \<sigma>" and "t \<cdot> \<delta> = t \<cdot> sop (-\<pi>\<^sub>1) \<circ>\<^sub>s \<sigma>"
      unfolding term_subst_eq_conv
      by (auto simp: \<delta>_def vars_rule_def subst_compose permute_atom_def) }
    then have conds1: "conds_sat R m cs\<^sub>1 \<delta>"
      using conds by (auto simp: conds_sat_iff) (auto simp: rename eqvt)
      
    { fix s t assume "(s, t) \<in> set cs\<^sub>2"
      then have "s \<cdot> \<delta> = s \<cdot> \<sigma>\<^sub>2" and "t \<cdot> \<delta> = t \<cdot> \<sigma>\<^sub>2"
        using * by (force simp: term_subst_eq_conv \<delta>_def vars_rule_def)+ }
    then have conds2': "conds_sat R n cs\<^sub>2 \<delta>" using conds2 by (auto simp: conds_sat_iff)
  
    from funposs have fp: "p <#> q \<in> funposs (clhs ((l\<^sub>1, r\<^sub>1), cs\<^sub>1))" by (simp add: rename)
    then have "p <#> q \<in> poss l\<^sub>1" by (auto dest: funposs_imp_poss)
    with vars_term_subt_at [OF this] and disj and l_\<tau>
      have "vars_term (l\<^sub>1 |_ (p <#> q)) \<inter> vars_term l\<^sub>2 = {}" and "(l\<^sub>1 |_ (p <#> q)) \<cdot> \<sigma>\<^sub>1 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2"
      apply (auto simp: eqvt vars_crule_def vars_rule_def rename)
      by (metis subt_at_eqvt term_pt.permute_minus_cancel(2))
    from vars_term_disjoint_imp_unifier [OF this]
      obtain \<mu> where "mgu (l\<^sub>1 |_ (p <#> q)) l\<^sub>2 = Some \<mu>"
      using mgu_complete by (auto simp: unifiers_def)
     with overlapI [OF rule_variants disj fp]
       have "overlap R ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) (p <#> q)" by auto
     with ao show ?thesis
       using funposs(1) and \<open>C \<noteq> MHole\<close> and conds1 and conds2' and comm
       by (cases; cases C) auto
  qed
qed

context
  fixes R :: "('f, 'v\<Colon>infinite) ctrs"
  assumes ao: "almost_orthogonal R"
    and po: "properly_oriented R"
    and rs: "right_stable R"
    and t3: "type3 R"
    and vars: "\<forall>((l, r), cs) \<in> R. is_Fun l"
begin

lemma trs_n_peak:
  assumes "(s, t) \<in> trs_n R (Suc m)" and "(s, u) \<in> trs_n R (Suc n)"
    and "\<And>m' n'. m' + n' < (Suc m) + (Suc n) \<Longrightarrow> comm ((epar_n R m')\<^sup>*) ((epar_n R n')\<^sup>*)"
  shows "\<exists>v. (t, v) \<in> epar_n R (Suc n) \<and> (u, v) \<in> epar_n R (Suc m)"
proof -
  from assms(1) obtain l\<^sub>1' r\<^sub>1' cs\<^sub>1' \<sigma>\<^sub>1' where s1: "s = l\<^sub>1' \<cdot> \<sigma>\<^sub>1'" and t: "t = r\<^sub>1' \<cdot> \<sigma>\<^sub>1'"
    and rule1': "((l\<^sub>1', r\<^sub>1'), cs\<^sub>1') \<in> R"
    and cs1': "\<forall>(s', t') \<in> set cs\<^sub>1'. (s' \<cdot> \<sigma>\<^sub>1', t' \<cdot> \<sigma>\<^sub>1') \<in> (rstep (trs_n R m))\<^sup>*"
    and conds1': "conds_sat R m cs\<^sub>1' \<sigma>\<^sub>1'"
      by (auto elim: trs_n_SucE simp: conds_sat_iff cstep_n_rstep_trs_n_conv)
  from assms(2) obtain l\<^sub>2 r\<^sub>2 cs\<^sub>2 \<sigma>\<^sub>2 where s2: "s = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" and u: "u = r\<^sub>2 \<cdot> \<sigma>\<^sub>2"
    and rule2: "((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R"
    and cs2: "\<forall>(s', t') \<in> set cs\<^sub>2. (s' \<cdot> \<sigma>\<^sub>2, t' \<cdot> \<sigma>\<^sub>2) \<in> (rstep (trs_n R n))\<^sup>*"
    and conds2: "conds_sat R n cs\<^sub>2 \<sigma>\<^sub>2"
      by (auto elim: trs_n_SucE simp: conds_sat_iff cstep_n_rstep_trs_n_conv)

  from vars_crule_disjoint obtain \<pi>\<^sub>1
    where \<pi>\<^sub>1: "vars_crule (\<pi>\<^sub>1 \<bullet> ((l\<^sub>1', r\<^sub>1'), cs\<^sub>1')) \<inter> vars_crule ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) = {}" ..
  def l\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> l\<^sub>1'" and r\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> r\<^sub>1'"
    and cs\<^sub>1 \<equiv> "\<pi>\<^sub>1 \<bullet> cs\<^sub>1'" and \<sigma>\<^sub>1 \<equiv> "sop (-\<pi>\<^sub>1) \<circ>\<^sub>s \<sigma>\<^sub>1'"
  note rename = l\<^sub>1_def r\<^sub>1_def cs\<^sub>1_def \<sigma>\<^sub>1_def

  have *: "-\<pi>\<^sub>1 \<bullet> ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<in> R" and "0 \<bullet> ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R"
    using rule1' and rule2 by (simp_all add: eqvt rename o_def)
  then have rule_variants: "\<exists>\<pi>. \<pi> \<bullet> ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<in> R" "\<exists>\<pi>. \<pi> \<bullet> ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<in> R" by blast+

  have disj: "vars_crule ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) \<inter> vars_crule ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) = {}"
    using \<pi>\<^sub>1 by (auto simp: eqvt rename)

  have "\<epsilon> \<in> funposs l\<^sub>1'" using rule1' and vars by (cases l\<^sub>1') (auto)
  then have funposs: "\<epsilon> \<in> funposs (clhs ((l\<^sub>1, r\<^sub>1), cs\<^sub>1))" by (simp add: rename)

  have "vars_term (l\<^sub>1 |_ \<epsilon>) \<inter> vars_term l\<^sub>2 = {}"
    and l: "(l\<^sub>1 |_ \<epsilon>) \<cdot> \<sigma>\<^sub>1 = l\<^sub>2 \<cdot> \<sigma>\<^sub>2" using s1 and s2 and disj
    by (auto simp: vars_crule_def vars_rule_def rename)
  from vars_term_disjoint_imp_unifier [OF this]
    obtain \<mu> where mgu: "mgu (l\<^sub>1 |_ \<epsilon>) l\<^sub>2 = Some \<mu>"
    using mgu_complete by (auto simp: unifiers_def)

  have "overlap R ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) ((l\<^sub>2, r\<^sub>2), cs\<^sub>2) \<epsilon>"
    using overlapI [OF rule_variants disj funposs] and mgu by auto
  with ao have "r\<^sub>1 \<cdot> \<mu> = r\<^sub>2 \<cdot> \<mu> \<or> (\<exists>\<pi>. \<pi> \<bullet> ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) = ((l\<^sub>2, r\<^sub>2), cs\<^sub>2)) \<or>
    (\<forall>m n. comm ((cstep_n R m)\<^sup>*) ((cstep_n R n)\<^sup>*) \<longrightarrow>
      \<not> (\<exists>\<sigma>. conds_sat R m cs\<^sub>1 \<sigma> \<and> conds_sat R n cs\<^sub>2 \<sigma>))"
    using mgu by (cases) (auto)
  moreover {
    def \<tau> \<equiv> "\<lambda>x. if x \<in> vars_crule ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) then \<sigma>\<^sub>1 x else \<sigma>\<^sub>2 x"
    have "l\<^sub>1 \<cdot> \<tau> = l\<^sub>1 \<cdot> \<sigma>\<^sub>1" and "l\<^sub>2 \<cdot> \<tau> = l\<^sub>2 \<cdot> \<sigma>\<^sub>2"
      and [simp]: "r\<^sub>1 \<cdot> \<tau> = r\<^sub>1 \<cdot> \<sigma>\<^sub>1" "r\<^sub>2 \<cdot> \<tau> = r\<^sub>2 \<cdot> \<sigma>\<^sub>2"
      using disj by (auto simp: term_subst_eq_conv vars_crule_def vars_rule_def \<tau>_def)
    then have "l\<^sub>1 \<cdot> \<tau> = l\<^sub>2 \<cdot> \<tau>" using l by simp
    from the_mgu [OF this] and mgu have "\<tau> = \<mu> \<circ>\<^sub>s \<tau>" by (auto simp: the_mgu_def)
    moreover assume "r\<^sub>1 \<cdot> \<mu> = r\<^sub>2 \<cdot> \<mu>"
    ultimately have "r\<^sub>1 \<cdot> \<tau> = r\<^sub>2 \<cdot> \<tau>" by (metis subst_compose_apply_term_distrib)
    then have "(t, t) \<in> epar_n R (Suc n) \<and> (u, t) \<in> epar_n R (Suc m)"
      by (simp add: t u) (simp add: rename)
    then have ?thesis ..
  }
  moreover {
    assume "\<forall>m n. comm ((cstep_n R m)\<^sup>*) ((cstep_n R n)\<^sup>*) \<longrightarrow>
      \<not> (\<exists>\<sigma>. conds_sat R m cs\<^sub>1 \<sigma> \<and> conds_sat R n cs\<^sub>2 \<sigma>)"
    moreover have "comm ((cstep_n R m)\<^sup>*) ((cstep_n R n)\<^sup>*)" using assms(3) [of m n]
      by (auto simp: rtrancl_epar_n_conv)
    ultimately have infeasible:
      "\<not> (\<exists>\<sigma>. conds_sat R m cs\<^sub>1 \<sigma> \<and> conds_sat R n cs\<^sub>2 \<sigma>)" by blast

    def \<delta> \<equiv> "\<lambda>x. if x \<in> \<Union>(vars_rule ` set cs\<^sub>1) then \<sigma>\<^sub>1' (-\<pi>\<^sub>1 \<bullet> x) else \<sigma>\<^sub>2 x"

    have *: "\<Union>(vars_rule ` set cs\<^sub>1) \<inter> \<Union>(vars_rule ` set cs\<^sub>2) = {}"
      using disj by (auto simp: vars_crule_def vars_trs_def)

    { fix s t assume "(s, t) \<in> set cs\<^sub>1"
      then have "s \<cdot> \<delta> = s \<cdot> sop (-\<pi>\<^sub>1) \<circ>\<^sub>s \<sigma>\<^sub>1'" and "t \<cdot> \<delta> = t \<cdot> sop (-\<pi>\<^sub>1) \<circ>\<^sub>s \<sigma>\<^sub>1'"
      unfolding term_subst_eq_conv
      by (auto simp: \<delta>_def vars_rule_def subst_compose permute_atom_def) }
      then have conds1: "conds_sat R m cs\<^sub>1 \<delta>"
        using conds1' by (auto simp: conds_sat_iff) (auto simp: rename eqvt)
      { fix s t assume "(s, t) \<in> set cs\<^sub>2"
        then have "s \<cdot> \<delta> = s \<cdot> \<sigma>\<^sub>2" and "t \<cdot> \<delta> = t \<cdot> \<sigma>\<^sub>2"
          using * by (force simp: term_subst_eq_conv \<delta>_def vars_rule_def)+ }
      then have conds2: "conds_sat R n cs\<^sub>2 \<delta>"
        using conds2 by (auto simp: conds_sat_iff)
      with infeasible and conds1 have ?thesis by blast
  }
  moreover {
    assume "\<exists>\<pi>. \<pi> \<bullet> ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) = ((l\<^sub>2, r\<^sub>2), cs\<^sub>2)"
    then obtain \<pi> where vt: "\<pi> \<bullet> ((l\<^sub>1, r\<^sub>1), cs\<^sub>1) = ((l\<^sub>2, r\<^sub>2), cs\<^sub>2)"
      using CCP_I [OF rule_variants disj funposs] and mgu by auto
    
    let ?\<sigma> = "(sop \<pi>) \<circ>\<^sub>s \<sigma>\<^sub>2"
    
    have \<sigma>: "\<forall>x \<in> vars_term l\<^sub>1. \<sigma>\<^sub>1 x = ?\<sigma> x" using vt l
      by (auto simp add: subst_compose permute_term_subst_apply_term term_subst_eq_conv eqvt)
  
    have ?thesis
    proof (cases "vars_term r\<^sub>1 \<subseteq> vars_term l\<^sub>1")
      case (True)
      then have "r\<^sub>1 \<cdot> \<sigma>\<^sub>1 = r\<^sub>1 \<cdot> ?\<sigma>" unfolding term_subst_eq_conv using \<sigma> by auto
      then have "(t, t) \<in> epar_n R (Suc n) \<and> (u, t) \<in> epar_n R (Suc m)"
        using vt u t r\<^sub>1_def \<sigma>\<^sub>1_def \<sigma> by (auto simp add: term_subst_eq_conv eqvt)
      then show ?thesis ..
    next
      case (False)
      let ?c = "\<lambda>i. take i cs\<^sub>1"
      let ?V = "\<lambda>i. \<Union>(vars_rule ` set (?c i))"
      let ?P1 = "\<lambda>\<rho>. \<forall>x \<in> vars_term l\<^sub>1. \<rho> x = \<sigma>\<^sub>1 x \<and> \<sigma>\<^sub>1 x = ?\<sigma> x"
      let ?P2 = "\<lambda>\<rho> i. subst_domain \<rho> \<subseteq> vars_term l\<^sub>1 \<union> ?V i"
      let ?P3 = "\<lambda>\<rho> i. \<forall>x \<in> vars_term l\<^sub>1 \<union> ?V i. (\<sigma>\<^sub>1 x, \<rho> x) \<in> (epar_n R n)\<^sup>* \<and> (?\<sigma> x, \<rho> x) \<in> (epar_n R m)\<^sup>*"
      { fix i :: nat
        assume "i \<le> length cs\<^sub>1"
        then have "\<exists>\<rho>. ?P1 \<rho> \<and> ?P2 \<rho> i \<and> ?P3 \<rho> i"
        proof (induction i)
          case (0)
          let ?\<rho> = "\<sigma>\<^sub>1 |s vars_term l\<^sub>1"
          have "?P1 ?\<rho>" using \<sigma> by simp
          moreover have "?P2 ?\<rho> 0" by (simp add: subst_domain_vars_term_subset)
          moreover have "?P3 ?\<rho> 0" using \<sigma> by (simp add: o_def)
          ultimately show ?case by blast
        next
          case (Suc i)
          then have len: "i < length cs\<^sub>1" and "i \<le> length cs\<^sub>1" by arith+
          from Suc(1) [OF this(2)]
            obtain \<rho> where a: "?P1 \<rho>" and b: "?P2 \<rho> i" and c: "?P3 \<rho> i" by blast
  
          let ?s\<^sub>i = "fst (cs\<^sub>1 ! i)" and ?t\<^sub>i = "snd (cs\<^sub>1 ! i)"
          let ?\<rho>\<^sub>1 = "((l\<^sub>1, r\<^sub>1), cs\<^sub>1)"
  
          from po and rule1' have "vars_term r\<^sub>1' \<subseteq> vars_term l\<^sub>1' \<or>
            (\<forall>i < length cs\<^sub>1'. vars_term (fst (cs\<^sub>1' ! i)) \<subseteq> X_vars ((l\<^sub>1', r\<^sub>1'), cs\<^sub>1') i)"
            by (force simp: properly_oriented_def)
          from permute_boolI [OF this, of \<pi>\<^sub>1, unfolded eqvt bool_pt.all_eqvt'] and False
          have "\<forall>i < length (snd ?\<rho>\<^sub>1). vars_term (fst (snd ?\<rho>\<^sub>1 ! i)) \<subseteq> X_vars ?\<rho>\<^sub>1 i"
            by (auto simp: l\<^sub>1_def r\<^sub>1_def cs\<^sub>1_def eqvt)
          with len have ssubs: "vars_term ?s\<^sub>i \<subseteq> vars_term l\<^sub>1 \<union> ?V i"
            by (auto simp: X_vars_def vars_rule_def)
          with c have "\<forall>x \<in> vars_term ?s\<^sub>i. (\<sigma>\<^sub>1 x, \<rho> x) \<in> (epar_n R n)\<^sup>* \<and> (?\<sigma> x, \<rho> x) \<in> (epar_n R m)\<^sup>*"
            by auto
          then have 1: "(?s\<^sub>i \<cdot> \<sigma>\<^sub>1, ?s\<^sub>i \<cdot> \<rho>) \<in> (epar_n R n)\<^sup>*" and 2: "(?s\<^sub>i \<cdot> ?\<sigma>, ?s\<^sub>i \<cdot> \<rho>) \<in> (epar_n R m)\<^sup>*"
            using subst_epar_n_imp_epar_n_rtrancl[of ?s\<^sub>i \<sigma>\<^sub>1 \<rho> R n]
            and subst_epar_n_imp_epar_n_rtrancl[of ?s\<^sub>i ?\<sigma> \<rho> R m] by auto
  
          from len vt have "i < length cs\<^sub>2" by (auto simp add: eqvt)
          with cs2 vt have "(?s\<^sub>i \<cdot> ?\<sigma>, ?t\<^sub>i \<cdot> ?\<sigma>) \<in> (rstep (trs_n R n))\<^sup>*" by (auto simp add: len eqvt)
          from rev_subsetD [OF this] have 3: "(?s\<^sub>i \<cdot> ?\<sigma>, ?t\<^sub>i \<cdot> ?\<sigma>) \<in> (epar_n R n)\<^sup>*"
            using cstep_n_rstep_trs_n_conv [of R n] and cstep_n_subset_epar_n [of R n]
            by (auto simp add: rtrancl_mono)
  
          from len cs\<^sub>1_def have len': "i < length cs\<^sub>1'" by auto
  
          from len' cs1' have "(?s\<^sub>i \<cdot> \<sigma>\<^sub>1, ?t\<^sub>i \<cdot> \<sigma>\<^sub>1) \<in> (rstep (trs_n R m))\<^sup>*"
            by (auto simp: cs\<^sub>1_def \<sigma>\<^sub>1_def eqvt)
          from rev_subsetD [OF this] have 4: "(?s\<^sub>i \<cdot> \<sigma>\<^sub>1, ?t\<^sub>i \<cdot> \<sigma>\<^sub>1) \<in> (epar_n R m)\<^sup>*"
            using cstep_n_rstep_trs_n_conv [of R m] and cstep_n_subset_epar_n [of R m]
            by (auto simp add: rtrancl_mono)
            
          from assms(3) have comm: "comm ((epar_n R m)\<^sup>*) ((epar_n R n)\<^sup>*)" by auto
          obtain u where 8: "(?t\<^sub>i \<cdot> \<sigma>\<^sub>1, u) \<in> (epar_n R n)\<^sup>*"
            and 5: "(?s\<^sub>i \<cdot> \<rho>, u) \<in> (epar_n R m)\<^sup>*"
            using commE [OF comm 4 1] by auto
          from 2 5 have 6: "(?s\<^sub>i \<cdot> ?\<sigma>, u) \<in> (epar_n R m)\<^sup>*" by auto
          have "cs\<^sub>1 ! i = (?s\<^sub>i, ?t\<^sub>i)" by auto
          then have cs: "take (Suc i) cs\<^sub>1 = take i cs\<^sub>1 @ [(?s\<^sub>i, ?t\<^sub>i)]"
            using len take_Suc_conv_app_nth by force
          obtain s' where 9: "(u, s') \<in> (epar_n R n)\<^sup>*"
            and 10: "(?t\<^sub>i \<cdot> ?\<sigma>, s') \<in> (epar_n R m)\<^sup>*"
            using commE [OF comm 6 3] by auto
          from 8 9 have 7: "(?t\<^sub>i \<cdot> \<sigma>\<^sub>1, s') \<in> (epar_n R n)\<^sup>*"  by auto
  
          from len' rule1' rs [unfolded right_stable_def, THEN bspec [of R _ "((l\<^sub>1', r\<^sub>1'), cs\<^sub>1')"]] have
            "linear_term (snd (cs\<^sub>1' ! i)) \<and> funas_term (snd (cs\<^sub>1' ! i)) \<subseteq> funas_ctrs R - { f. defined (Ru R) f } \<or>
             ground (snd (cs\<^sub>1' ! i)) \<and> (snd (cs\<^sub>1' ! i)) \<in> NF (rstep (Ru R))"
             by (simp add: Let_def)
          then have
            "linear_term ?t\<^sub>i \<and> funas_term ?t\<^sub>i \<subseteq> funas_ctrs R - { f. defined (Ru R) f } \<or>
             ground ?t\<^sub>i \<and> ?t\<^sub>i \<in> NF (rstep (Ru R))"
             using len' and rule1'
             by (auto simp only: l\<^sub>1_def r\<^sub>1_def cs\<^sub>1_def rules_pt.nth_eqvt [symmetric] eqvt [symmetric]
               linear_term_permute ground_imp_linear_term funas_term_permute ground_permute
               funas_ctrs_def funas_crule_def [abs_def] funas_trs_def funas_rule_def) (auto)
          then show ?case
          proof
            assume gnf: "ground ?t\<^sub>i \<and> ?t\<^sub>i \<in> NF (rstep (Ru R))"
            then have "?t\<^sub>i \<cdot> \<sigma>\<^sub>1 = ?t\<^sub>i" by (simp add: ground_subst_apply)
            with gnf have "?t\<^sub>i \<cdot> \<sigma>\<^sub>1 \<in> NF (rstep (Ru R))" by auto
            with Ru_NF_imp_R_NF have "?t\<^sub>i \<cdot> \<sigma>\<^sub>1 \<in> NF (cstep R)" by auto
            with 7 have eq1: "?t\<^sub>i \<cdot> \<sigma>\<^sub>1 = s'"
              using epars_n_subset_csteps_n
              by (metis (no_types, hide_lams) NF_not_suc contra_subsetD cstep_iff prod.collapse rtrancl_mono subsetI)
            from gnf have "?t\<^sub>i \<cdot> ?\<sigma> = ?t\<^sub>i" by (metis ground_subst_apply)
            with gnf have "?t\<^sub>i \<cdot> ?\<sigma> \<in> NF (rstep (Ru R))" by auto
            with Ru_NF_imp_R_NF have "?t\<^sub>i \<cdot> ?\<sigma> \<in> NF (cstep R)" by auto
            with 10 have eq2: "?t\<^sub>i \<cdot> ?\<sigma> = s'"
              using epars_n_subset_csteps_n
              by (metis (no_types, hide_lams) NF_not_suc contra_subsetD cstep_iff prod.collapse rtrancl_mono subsetI)
            from eq1 eq2 have "?t\<^sub>i \<cdot> \<sigma>\<^sub>1 = ?t\<^sub>i \<cdot> ?\<sigma>" ..
            show ?thesis
            proof (intro exI[of _ \<rho>] conjI)
              show "?P1 \<rho>" by (fact a)
            next
              show "?P2 \<rho> (Suc i)" using b unfolding take_Suc_conv_app_nth[OF len] by force
              
            next
              from gnf have "ground ?t\<^sub>i" ..
              then have nov: "vars_term ?t\<^sub>i = {}" by (simp add: ground_vars_term_empty)
              with ssubs have x:
                "vars_rule (?s\<^sub>i, ?t\<^sub>i) \<subseteq> vars_term l\<^sub>1 \<union> \<Union>(vars_rule ` set (take i cs\<^sub>1))"
                by (simp add: vars_rule_def)              
              from cs x have
                "vars_term l\<^sub>1 \<union> \<Union>(vars_rule ` set (take i cs\<^sub>1)) = vars_term l\<^sub>1 \<union> \<Union>(vars_rule ` set (take (Suc i) cs\<^sub>1))"
                by auto
              then show "?P3 \<rho> (Suc i)" using c by auto
            qed
          next
            assume lcd: "linear_term ?t\<^sub>i \<and> funas_term ?t\<^sub>i \<subseteq> funas_ctrs R - { f. defined (Ru R) f }"
            from lcd have lin: "linear_term ?t\<^sub>i"
              and cterm: "funas_term ?t\<^sub>i \<subseteq> funas_ctrs R - { f. defined (Ru R) f }" by auto
  
            from 10 have "(snd (cs\<^sub>1 ! i) \<cdot> ?\<sigma>, s') \<in> (cstep_n R m)\<^sup>*"
              by (meson epar_n_subset_csteps_n rtrancl_subset_rtrancl subset_eq)
            with lin have "\<exists> \<delta>. ?t\<^sub>i \<cdot> \<delta> = s' \<and> subst_domain \<delta> \<subseteq> vars_term ?t\<^sub>i \<and>
              (\<forall>x \<in> vars_term ?t\<^sub>i. (?\<sigma> x, \<delta> x) \<in> (cstep_n R m)\<^sup>*)"
            proof (cases rule: linear_term_rtrancl_cstep_n_cases)
              case (varposs \<tau>)
              then show ?thesis by blast
            next
              case (funposs \<tau> p u)
              from no_step_from_constructor [OF vars cterm funposs(2), of \<tau> u m]
              show ?thesis using funposs(3) ..
            qed
            then obtain \<gamma> where "?t\<^sub>i \<cdot> \<gamma> = s'" and dom: "subst_domain \<gamma> \<subseteq> vars_term ?t\<^sub>i" 
              and "\<forall>x \<in> vars_term ?t\<^sub>i. (?\<sigma> x, \<gamma> x) \<in> (cstep_n R m)\<^sup>*" by blast
            then have *: "\<forall>x \<in> vars_term ?t\<^sub>i. (?\<sigma> x, \<gamma> x) \<in> (epar_n R m)\<^sup>*"
              using rtrancl_mono[OF cstep_n_subset_epar_n] by blast
  
            from 7 have 11: "(?t\<^sub>i \<cdot> \<sigma>\<^sub>1, s') \<in> (cstep_n R n)\<^sup>*"
              by (meson epar_n_subset_csteps_n rtrancl_subset_rtrancl subset_eq)
            from lin 11 have "\<exists> \<delta>. ?t\<^sub>i \<cdot> \<delta> = s' \<and> subst_domain \<delta> \<subseteq> vars_term ?t\<^sub>i \<and>
              (\<forall>x \<in> vars_term ?t\<^sub>i. (\<sigma>\<^sub>1 x, \<delta> x) \<in> (cstep_n R n)\<^sup>*)"
            proof (cases rule: linear_term_rtrancl_cstep_n_cases)
              case (varposs \<tau>)
              then show ?thesis by blast
            next
              case (funposs \<tau> p u)
              from no_step_from_constructor [OF vars cterm funposs(2), of \<tau> u n]
              show ?thesis using funposs(3) ..
            qed
            then obtain \<delta> where "?t\<^sub>i \<cdot> \<delta> = s'" and dom: "subst_domain \<delta> \<subseteq> vars_term ?t\<^sub>i" 
              and "\<forall>x \<in> vars_term ?t\<^sub>i. (\<sigma>\<^sub>1 x, \<delta> x) \<in> (cstep_n R n)\<^sup>*" by blast
            then have **: "\<forall>x \<in> vars_term ?t\<^sub>i. (\<sigma>\<^sub>1 x, \<delta> x) \<in> (epar_n R n)\<^sup>*"
              using rtrancl_mono[OF cstep_n_subset_epar_n] by blast
  
            from len' rule1' rs [unfolded right_stable_def, THEN bspec [of R _ "((l\<^sub>1', r\<^sub>1'), cs\<^sub>1')"]]
              have
              "(vars_term l\<^sub>1' \<union> \<Union>(vars_term ` lhss (set (take (Suc i) cs\<^sub>1'))) \<union>
               \<Union>(vars_term ` rhss (set (take i cs\<^sub>1')))) \<inter> vars_term (snd (cs\<^sub>1' ! i)) = {}"
              by (simp add: Let_def)
            then have inter:
              "(vars_term l\<^sub>1 \<union> \<Union>(vars_term ` lhss (set (?c (Suc i)))) \<union>
              \<Union>(vars_term ` rhss (set (?c i)))) \<inter> vars_term ?t\<^sub>i = {}"
              using len' rule1' by (simp add: l\<^sub>1_def cs\<^sub>1_def eqvt [symmetric] take_map)
            
            from inter have disj2: "vars_term l\<^sub>1 \<inter> vars_term (snd (cs\<^sub>1 ! i)) = {}" by auto
            have "\<Union>(vars_rule ` set (take i cs\<^sub>1)) = \<Union>(vars_term ` lhss (set (take i cs\<^sub>1))) \<union> \<Union>(vars_term ` rhss (set (take i cs\<^sub>1)))"
              by (auto simp add: vars_rule_def)
            also have "... \<subseteq> \<Union>(vars_term ` lhss (set (take (Suc i) cs\<^sub>1))) \<union> \<Union>(vars_term ` rhss (set (take i cs\<^sub>1)))"
              using len take_Suc_conv_app_nth by fastforce
            finally have "\<Union>(vars_rule ` set (take i cs\<^sub>1)) \<subseteq> \<Union>(vars_term ` lhss (set (take (Suc i) cs\<^sub>1))) \<union> \<Union>(vars_term ` rhss (set (take i cs\<^sub>1)))"
              by auto
            then have "vars_term l\<^sub>1 \<union> \<Union>(vars_rule ` set (take i cs\<^sub>1)) \<subseteq> (vars_term l\<^sub>1 \<union> \<Union>(vars_term ` lhss (set (take (Suc i) cs\<^sub>1))) \<union> \<Union>(vars_term ` rhss (set (take i cs\<^sub>1))))"
              by auto
            with b have "subst_domain \<rho> \<subseteq> (vars_term l\<^sub>1 \<union> \<Union>(vars_term ` lhss (set (take (Suc i) cs\<^sub>1))) \<union> \<Union>(vars_term ` rhss (set (take i cs\<^sub>1))))"
              by (auto) (metis (no_types, lifting) UN_E UnE subsetCE)
            with inter dom have disj: "subst_domain \<delta> \<inter> subst_domain \<rho> = {}" by auto
            from dom inter have lid: "\<forall>x \<in> vars_term l\<^sub>1. \<delta> x = Var x" unfolding subst_domain_def by auto
            let ?\<rho>\<^sub>i = "\<lambda>x. (if x \<in> vars_term ?t\<^sub>i then \<delta> x else \<rho> x)"
            show ?thesis
            proof (intro exI[of _ ?\<rho>\<^sub>i] conjI)
              from lid disj2 have "\<forall>x \<in> vars_term l\<^sub>1. ?\<rho>\<^sub>i x = \<rho> x" by auto
              then show "?P1 ?\<rho>\<^sub>i" using a by auto
            next
              from dom have "subst_domain \<delta> \<subseteq> vars_term (snd (cs\<^sub>1 ! i))" by auto
              also have "\<dots> \<subseteq> \<Union>(vars_rule ` set (take (Suc i) cs\<^sub>1))" using cs
                by (auto) (simp add: vars_rule_def)
              also have "\<dots> \<subseteq> vars_term l\<^sub>1 \<union> \<Union>(vars_rule ` set (take (Suc i) cs\<^sub>1))" by auto
              finally have "subst_domain \<delta> \<subseteq> vars_term l\<^sub>1 \<union> \<Union>(vars_rule ` set (take (Suc i) cs\<^sub>1))" by auto
              moreover have "subst_domain \<rho> \<subseteq> vars_term l\<^sub>1 \<union> \<Union>(vars_rule ` set (take (Suc i) cs\<^sub>1))"
                using b cs by auto
              ultimately show "?P2 ?\<rho>\<^sub>i (Suc i)"
                by (auto split: if_splits simp add: subst_domain_def)
            next
              { fix x
                assume "x \<in> vars_term l\<^sub>1 \<union> \<Union>(vars_rule ` set (take (Suc i) cs\<^sub>1))"
                then have "x \<in> vars_term l\<^sub>1 \<union> \<Union>(vars_rule ` set (take i cs\<^sub>1)) \<or> x \<in> vars_term ?t\<^sub>i"
                  using ssubs len unfolding cs by (auto simp add: vars_rule_def)
                then have "(\<sigma>\<^sub>1 x, ?\<rho>\<^sub>i x) \<in> (epar_n R n)\<^sup>* \<and> ((?\<sigma>) x, ?\<rho>\<^sub>i x) \<in> (epar_n R m)\<^sup>*"
                proof
                  assume "x \<in> vars_term l\<^sub>1 \<union> \<Union>(vars_rule ` set (take i cs\<^sub>1))"
                  moreover with c have "(\<sigma>\<^sub>1 x, \<rho> x) \<in> (epar_n R n)\<^sup>*"
                    and "(?\<sigma> x, \<rho> x) \<in> (epar_n R m)\<^sup>*" by auto
                  ultimately show ?thesis using inter unfolding cs
                    by (auto simp add: vars_rule_def)
                next
                  assume "x \<in> vars_term ?t\<^sub>i"
                  moreover have "\<forall>x \<in> vars_term ?t\<^sub>i. \<delta> x = \<gamma> x"
                    using \<open>?t\<^sub>i \<cdot> \<gamma> = s'\<close> \<open>?t\<^sub>i \<cdot> \<delta> = s'\<close>
                    by (auto simp add: term_subst_eq_conv)
                  ultimately show ?thesis using ** * by simp
                qed
              }
              then show "?P3 ?\<rho>\<^sub>i (Suc i)" ..
            qed
          qed
        qed }
        from this [OF le_refl] obtain \<rho>
          where "?P1 \<rho>" and "?P2 \<rho> (length cs\<^sub>1)" and C: "?P3 \<rho> (length cs\<^sub>1)" by auto
        have "vars_term r\<^sub>1' \<subseteq> vars_term l\<^sub>1' \<union> \<Union>(vars_rule ` set cs\<^sub>1')"
          using t3 and rule1' by (force simp: type3_def vars_trs_def)
        from permute_boolI [OF this, of \<pi>\<^sub>1, unfolded eqvt]
          have "vars_term r\<^sub>1 \<subseteq> vars_term l\<^sub>1 \<union> \<Union>(vars_rule ` set cs\<^sub>1)"
          by (auto simp add: l\<^sub>1_def r\<^sub>1_def cs\<^sub>1_def eqvt)
        with C [simplified] and subst_epar_n_imp_epar_n_rtrancl [of r\<^sub>1 _ _ R]
          have "(r\<^sub>1 \<cdot> \<sigma>\<^sub>1, r\<^sub>1 \<cdot> \<rho>) \<in> (epar_n R n)\<^sup>*"
          and "(r\<^sub>1 \<cdot> ?\<sigma>, r\<^sub>1 \<cdot> \<rho>) \<in> (epar_n R m)\<^sup>*" by blast+
        moreover have "u = \<pi> \<bullet> r\<^sub>1 \<cdot> \<sigma>\<^sub>2" using vt by (simp add: u eqvt)
        ultimately have "(t, r\<^sub>1 \<cdot> \<rho>) \<in> epar_n R (Suc n)"
          and "(u, r\<^sub>1 \<cdot> \<rho>) \<in> epar_n R (Suc m)"
          using epar_n_rtrancl_Suc_mono
          by (auto simp: t r\<^sub>1_def \<sigma>\<^sub>1_def)
        then show ?thesis by blast
    qed
  }
  ultimately show ?thesis by blast
qed

lemma epar_n_varpeak:
  assumes s: "s =\<^sub>f (C, ts)" and ne: "C \<noteq> MHole" and len: "length us = length ts"
    and trs_m: "(s, t) \<in> trs_n R (Suc m)"
    and *: "\<forall>i < length ts. (ts ! i, us ! i) \<in> trs_n R (Suc n) \<union> (cstep_n R n)\<^sup>*"
    and comm1: "\<And>m' n'. m' + n' < (Suc m) + (Suc n) \<Longrightarrow> comm ((epar_n R m')\<^sup>*) ((epar_n R n'))"
    and comm2: "\<And>m' n'. m' + n' < (Suc m) + (Suc n) \<Longrightarrow> comm ((epar_n R m')\<^sup>*) ((epar_n R n')\<^sup>*)"
  shows "\<exists>v. (t, v) \<in> epar_n R (Suc n) \<and> (fill_holes C us, v) \<in> epar_n R (Suc m)"
proof -
  obtain l r \<sigma> cs
    where s': "s = l \<cdot> \<sigma>" and t: "t = r \<cdot> \<sigma>"
    and rule: "((l, r), cs) \<in> R"
    and ***: "\<forall>(s\<^sub>i, t\<^sub>i) \<in> set cs. (s\<^sub>i \<cdot> \<sigma>, t\<^sub>i \<cdot> \<sigma>) \<in> (rstep (trs_n R m))\<^sup>*"
    by (rule trs_n_SucE [OF trs_m])
  then have l_\<sigma>: "l \<cdot> \<sigma> =\<^sub>f (C, ts)" and conds: "conds_sat R m cs \<sigma>"
    using s by (auto simp: conds_sat_iff cstep_n_rstep_trs_n_conv)
  from comm2 [of m n] have comm: "comm ((cstep_n R m)\<^sup>*) ((cstep_n R n)\<^sup>*)"
    by (auto simp: rtrancl_epar_n_conv)
  from rule and ao have "linear_term l" by simp
  from epar_n_varpeak' [OF comm ao rule conds l_\<sigma> ne * len] obtain \<tau>
    where l_\<tau>: "l \<cdot> \<tau> =\<^sub>f (C, us)"
    and "subst_domain \<tau> \<subseteq> vars_term l"
    and **: "\<forall>x \<in> vars_term l. (\<sigma> x, \<tau> x) \<in> epar_n R (Suc n)" by blast

  let ?r = "((l, r), cs)"
  let ?c = "\<lambda>i. take i cs"
  let ?V = "\<lambda>i. \<Union>(vars_rule ` set (?c i))"
  let ?P1 = "\<lambda>\<rho>. \<forall>x \<in> vars_term l. \<rho> x = \<tau> x"
  let ?P2 = "\<lambda>\<rho> i. subst_domain \<rho> \<subseteq> vars_term l \<union> ?V i"
  let ?P3 = "\<lambda>\<rho> i. conds_sat R m (?c i) \<rho>"
  let ?P4 = "\<lambda>\<rho> i. \<forall>x \<in> vars_term l \<union> ?V i. (\<sigma> x, \<rho> x) \<in> epar_n R (Suc n)"
  { fix i :: nat
    assume "i \<le> length cs"
    then have "\<exists>\<rho>. ?P1 \<rho> \<and> ?P2 \<rho> i \<and> ?P3 \<rho> i \<and> ?P4 \<rho> i"
    proof (induction i)
      case 0
      let ?\<rho> = "\<tau> |s vars_term l"
      have "?P1 ?\<rho>" by simp
      moreover have "?P2 ?\<rho> 0" by (simp add: subst_domain_vars_term_subset)
      moreover have "?P3 ?\<rho> 0" by (simp add: conds_sat_iff)
      moreover have "?P4 ?\<rho> 0" using ** by (simp add: o_def)
      ultimately show ?case by blast
    next
      case (Suc i)
      then have len: "i < length cs" and "i \<le> length cs" by arith+
      from Suc(1) [OF this(2)]
        obtain \<rho>' where a: "?P1 \<rho>'" and b: "?P2 \<rho>' i" and c: "?P3 \<rho>' i" and d: "?P4 \<rho>' i" by blast

      let ?s\<^sub>i = "fst (cs ! i)" and ?t\<^sub>i = "snd (cs ! i)"

      from len *** have "(?s\<^sub>i \<cdot> \<sigma>, ?t\<^sub>i \<cdot> \<sigma>) \<in> (cstep_n R m)\<^sup>*"
        by (auto simp add: case_prod_beta cstep_n_rstep_trs_n_conv)
      then have 1: "(?s\<^sub>i \<cdot> \<sigma>, ?t\<^sub>i \<cdot> \<sigma>) \<in> (epar_n R m)\<^sup>*"
        by (metis cstep_n_subset_epar_n epar_n_subset_csteps_n rtrancl_subset)

      let ?W = "vars_term ?s\<^sub>i - (vars_term l \<union> ?V i)"
      let ?\<rho> = "\<lambda>x. (if x \<in> ?W then (\<sigma> |s ?W) x else \<rho>' x)"

      have disj: "?W \<inter> (vars_term l \<union> ?V i) = {}" by (auto simp add: vars_rule_def)
      with d have w: "\<forall>x \<in> vars_term ?s\<^sub>i. (\<sigma> x, ?\<rho> x) \<in> (epar_n R (Suc n))" by auto
      then have 2: "(?s\<^sub>i \<cdot> \<sigma>, ?s\<^sub>i \<cdot> ?\<rho>) \<in> (epar_n R (Suc n))" by (rule subst_epar_n_imp_epar_n)

      from comm1 have comm: "comm ((epar_n R m)\<^sup>*) ((epar_n R (Suc n)))" by auto
      obtain t' where 3: "(?t\<^sub>i \<cdot> \<sigma>, t') \<in> (epar_n R (Suc n))" and 4: "(?s\<^sub>i \<cdot> ?\<rho>, t') \<in> (epar_n R m)\<^sup>*"
        using commE [OF comm 1 2] by auto
      have cs: "take (Suc i) cs = take i cs @ [(?s\<^sub>i, ?t\<^sub>i)]" using len take_Suc_conv_app_nth by force
      from rs [unfolded right_stable_def, THEN bspec, OF rule] have
        "linear_term ?t\<^sub>i \<and> funas_term ?t\<^sub>i \<subseteq> funas_ctrs R - { f. defined (Ru R) f } \<or>
         ground ?t\<^sub>i \<and> ?t\<^sub>i \<in> NF (rstep (Ru R))" using len by (metis snd_conv)
      then show ?case
      proof (cases rule: disjE [consumes 1, case_names lct gnf])
        case (gnf)
        then have eq2: "?t\<^sub>i \<cdot> \<sigma> = ?t\<^sub>i" by (simp add: ground_subst_apply)
        with gnf have "?t\<^sub>i \<cdot> \<sigma> \<in> NF (rstep (Ru R))" by auto
        with Ru_NF_imp_R_NF have "?t\<^sub>i \<cdot> \<sigma> \<in> NF (cstep R)" by auto
        with 3 have eq1: "?t\<^sub>i \<cdot> \<sigma> = t'" using epar_n_subset_csteps_n
          by (metis (no_types, hide_lams) NF_not_suc contra_subsetD cstep_iff prod.collapse rtrancl_mono subsetI)
        from eq1 eq2 have "?t\<^sub>i = t'" by auto
        then have t': "t' = ?t\<^sub>i \<cdot> ?\<rho>" using gnf by (simp add: ground_subst_apply)
        from gnf have [simp]: "vars_term ?t\<^sub>i = {}" by (simp add: ground_vars_term_empty)
        from po [unfolded properly_oriented_def]
        show ?thesis
        proof (intro exI[of _ ?\<rho>] conjI)
          show "?P1 ?\<rho>" using a by auto
        next
          show "?P2 ?\<rho> (Suc i)" using b gnf unfolding take_Suc_conv_app_nth[OF len]
            by (auto simp add: subst_domain_def vars_rule_def)
        next
          { fix s t
            assume *: "(s, t) \<in> set (take i cs)"
            with c have "(s \<cdot> \<rho>', t \<cdot> \<rho>') \<in> (cstep_n R m)\<^sup>*" by (auto simp: conds_sat_iff)
            moreover have "s \<cdot> ?\<rho> = s \<cdot> \<rho>'" and "t \<cdot> ?\<rho> = t \<cdot> \<rho>'"
              using * by (auto simp: vars_rule_def term_subst_eq_conv)
            ultimately have "(s \<cdot> ?\<rho>, t \<cdot> ?\<rho>) \<in> (cstep_n R m)\<^sup>*" by auto }
          then have "conds_sat R m (take i cs) ?\<rho>" by (auto simp add: conds_sat_iff)
          then show "?P3 ?\<rho> (Suc i)"
            using 4 len rtrancl_mono [OF epar_n_subset_csteps_n, of R m]
            by (auto simp add: t' conds_sat_iff cs)
        next
         show "?P4 ?\<rho> (Suc i)" using d 2 by (auto simp add: vars_rule_def cs)
        qed
      next
        case (lct)
        then have lin: "linear_term ?t\<^sub>i"
            and cterm: "funas_term ?t\<^sub>i \<subseteq> funas_ctrs R - { f. defined (Ru R) f }" by auto
        from lin 3 have "\<exists> \<delta>. ?t\<^sub>i \<cdot> \<delta> = t' \<and> subst_domain \<delta> \<subseteq> vars_term ?t\<^sub>i \<and>
          (\<forall>x \<in> vars_term ?t\<^sub>i. (\<sigma> x, \<delta> x) \<in> epar_n R (Suc n))"
        proof (cases rule: linear_term_epar_n_cases)
          case (varposs \<tau>)
          then show ?thesis by blast
        next
          case (funposs \<tau> u p)
          from no_step_from_constructor [OF vars cterm funposs(1), of \<tau> u "Suc n"]
          show ?thesis using funposs(2) ..
        qed
        then obtain \<gamma> where "?t\<^sub>i \<cdot> \<gamma> = t'" and dom: "subst_domain \<gamma> \<subseteq> vars_term ?t\<^sub>i" 
          and ***: "\<forall>x \<in> vars_term ?t\<^sub>i. (\<sigma> x, \<gamma> x) \<in> epar_n R (Suc n)" by blast

        have "\<Union>(vars_term ` lhss (set (take i cs))) \<union> \<Union>(vars_term ` rhss (set (take i cs))) = ?V i"
            by (auto simp add: vars_rule_def)
        then have "\<Union>(vars_term ` lhss (set (take (Suc i) cs))) \<union>
          \<Union>(vars_term ` rhss (set (take i cs))) = ?V i \<union> vars_term ?s\<^sub>i"
          unfolding cs by auto
        moreover from rs rule len have "(vars_term (clhs ?r) \<union>
          \<Union>(vars_term ` lhss (set (take (Suc i) (snd ?r)))) \<union>
          \<Union>(vars_term ` rhss (set (take i (snd ?r))))) \<inter> vars_term ?t\<^sub>i = {}"
          unfolding right_stable_def by (metis snd_conv)
        ultimately have inter: "(vars_term l \<union> ?V i \<union> vars_term ?s\<^sub>i) \<inter> vars_term ?t\<^sub>i = {}" by (auto)
        then have disj2: "vars_term l \<inter> vars_term ?t\<^sub>i = {}" by auto
        from dom inter have lid: "\<forall>x \<in> vars_term l. \<gamma> x = Var x" unfolding subst_domain_def by auto
        let ?\<mu> = "\<lambda>x. (if x \<in> vars_term ?t\<^sub>i then \<gamma> x else ?\<rho> x)"

        show ?thesis
        proof (intro exI[of _ ?\<mu>] conjI)
          from lid disj2 have "\<forall>x \<in> vars_term l. ?\<mu> x = ?\<rho> x" by auto
          then show "?P1 ?\<mu>" using a by auto
        next
          from dom have "subst_domain \<gamma> \<subseteq> vars_term (snd (cs ! i))" by auto
          also have "\<dots> \<subseteq> \<Union>(vars_rule ` set (take (Suc i) cs))" using cs
            by (auto) (simp add: vars_rule_def)
          also have "\<dots> \<subseteq> vars_term l \<union> \<Union>(vars_rule ` set (take (Suc i) cs))" by auto
          finally have "subst_domain \<gamma> \<subseteq> vars_term l \<union> \<Union>(vars_rule ` set (take (Suc i) cs))" by auto
          moreover have "subst_domain ?\<mu> \<subseteq> vars_term l \<union> \<Union>(vars_rule ` set (take (Suc i) cs))"
            using b cs by (auto simp add: subst_domain_def vars_rule_def)
          ultimately show "?P2 ?\<mu> (Suc i)" by (auto split: if_splits simp add: subst_domain_def)
        next
          { fix s t
            assume *: "(s, t) \<in> set (take i cs)"
            with c have "(s \<cdot> \<rho>', t \<cdot> \<rho>') \<in> (cstep_n R m)\<^sup>*" by (auto simp: conds_sat_iff)
            moreover have "s \<cdot> ?\<mu> = s \<cdot> \<rho>'" and "t \<cdot> ?\<mu> = t \<cdot> \<rho>'"
              using * inter len by (fastforce simp: vars_rule_def term_subst_eq_conv)+
            ultimately have "(s \<cdot> ?\<mu>, t \<cdot> ?\<mu>) \<in> (cstep_n R m)\<^sup>*" by auto }
          then have [simp]: "conds_sat R m (take i cs) \<rho>' \<Longrightarrow> conds_sat R m (take i cs) ?\<mu>"
            unfolding conds_sat_iff by (auto)
          have *: "?s\<^sub>i \<cdot> ?\<mu> = ?s\<^sub>i \<cdot> ?\<rho>" "?t\<^sub>i \<cdot> ?\<mu> = ?t\<^sub>i \<cdot> \<gamma>"
            using inter by (auto simp add: term_subst_eq_conv vars_rule_def)
          from \<open>?t\<^sub>i \<cdot> \<gamma> = t'\<close> have "t' = ?t\<^sub>i \<cdot> \<gamma>" ..
          with 4 have "(?s\<^sub>i \<cdot> ?\<rho>, ?t\<^sub>i \<cdot> \<gamma>) \<in> (epar_n R m)\<^sup>*" by auto
          then have "(?s\<^sub>i \<cdot> ?\<mu>, ?t\<^sub>i \<cdot> ?\<mu>) \<in> (cstep_n R m)\<^sup>*"
            using * and epars_n_subset_csteps_n [of R m]
            by (auto simp add: vars_rule_def)
          with c show "?P3 ?\<mu> (Suc i)" by (simp add: cs)
        next
          show "?P4 ?\<mu> (Suc i)" using d *** unfolding cs by (auto simp add: vars_rule_def)
        qed
      qed
    qed }
  from this [OF le_refl]
    obtain \<rho> where A: "?P1 \<rho>" and "?P2 \<rho> (length cs)"
    and C: "?P3 \<rho> (length cs)" and D: "?P4 \<rho> (length cs)"
    by blast
  have "l \<cdot> \<tau> = l \<cdot> \<rho>" using A by (simp add: term_subst_eq_conv)
  moreover have "conds_sat R m cs \<rho>" using C by simp
  ultimately have "(l \<cdot> \<tau>, r \<cdot> \<rho>) \<in> cstep_n R (Suc m)"
    using rule apply (auto simp: conds_sat_iff)
    by (metis cstep_n_rstep_trs_n_conv rstep.id splitD trs_n_SucI)
  then have "(fill_holes C us, r \<cdot> \<rho>) \<in> epar_n R (Suc m)"
    using cstep_n_subset_epar_n eqfE(1) l_\<tau> by fastforce
  moreover
  have "(t, r \<cdot> \<rho>) \<in> epar_n R (Suc n)"
  proof -
    have "vars_term r \<subseteq> vars_term l \<union> \<Union>(vars_rule ` set cs)"
      using t3 and rule by (force simp: type3_def vars_trs_def vars_rule_def)
    then show ?thesis using D by (auto simp: t subset_iff subst_epar_n_imp_epar_n)
  qed
  ultimately
  show ?thesis by blast
qed

lemma comm_epar_n:
  assumes "(s, t) \<in> epar_n R m" and "(s, u) \<in> epar_n R n"
  shows "\<exists>v. (t, v) \<in> epar_n R n \<and> (u, v) \<in> epar_n R m"
using assms
proof (induct "m + n" arbitrary: s t u m n rule: less_induct)
  case (less)

  have 1: "\<And>m' n'. m' + n' < m + n \<Longrightarrow> comm (epar_n R m') (epar_n R n')"
    using less.hyps by (auto simp: comm_def)
  have 2: "\<And>m' n'. m' + n' < m + n \<Longrightarrow> comm ((epar_n R m')\<^sup>*) (epar_n R n')"
    by (rule comm_rtrancl) fact
  have 3: "\<And>m' n'. m' + n' < m + n \<Longrightarrow> comm ((epar_n R m')\<^sup>*) ((epar_n R n')\<^sup>*)"
    by (subst comm_swap, rule comm_rtrancl, subst comm_swap) fact

  { assume "m = 0 \<or> n = 0"
    then have ?case using less.prems by auto }
  moreover
  { fix m' and n'
    assume [simp]: "m = Suc m'" "n = Suc n'"
    from less.prems obtain C ss ts and D us vs
      where s: "s =\<^sub>f (C, ss)" and t: "t =\<^sub>f (C, ts)" and [simp]: "length ts = length ss"
        and ss: "\<forall>i < length ss. (ss ! i, ts ! i) \<in> trs_n R m \<union> (cstep_n R m')\<^sup>*"
        and s': "s =\<^sub>f (D, us)" and u: "u =\<^sub>f (D, vs)" and [simp]: "length vs = length us"
        and us: "\<forall>i < length us. (us ! i, vs ! i) \<in> trs_n R n \<union> (cstep_n R n')\<^sup>*"
        by (auto elim!: epar_n_SucE)
    then have ss_epar: "\<forall>i < length ss. (ss ! i, ts ! i) \<in> epar_n R m"
      and us_epar: "\<forall>i < length us. (us ! i, vs ! i) \<in> epar_n R n"
      using trs_n_subset_epar_n [of R] and csteps_n_subset_epar_n_Suc [of R]
      by (simp_all) (blast)+
    
    have [simp]: "num_holes C = length ss" "num_holes D = length us"
      using t and u by (auto dest: eqfE)
    
    def E \<equiv> "C \<sqinter> D"
    def Cs \<equiv> "inf_mctxt_args C D"
    def Ds \<equiv> "inf_mctxt_args D C"
    have len_Cs: "num_holes E = length Cs" and C: "C = fill_holes_mctxt E Cs"
      and len_Ds: "num_holes E = length Ds" and D: "D = fill_holes_mctxt E Ds"
      by (simp_all add: E_def Cs_def Ds_def num_holes_inf_mctxt
           inf_mctxt_inf_mctxt_args inf_mctxt_inf_mctxt_args2)
         (metis inf_commute num_holes_inf_mctxt)
    have s_left: "s = fill_holes (fill_holes_mctxt E Cs) ss"
      and [simp]: "t = fill_holes (fill_holes_mctxt E Cs) ts"
      and s_right: "s = fill_holes (fill_holes_mctxt E Ds) us"
      and [simp]: "u = fill_holes (fill_holes_mctxt E Ds) vs"
      using s and s' and t and u by (auto simp: C D dest: eqfE)

    def [simp]: sss \<equiv> "partition_holes ss Cs"
    def [simp]: tss \<equiv> "partition_holes ts Cs"
    def [simp]: uss \<equiv> "partition_holes us Ds"
    def [simp]: vss \<equiv> "partition_holes vs Ds"

    def ss' \<equiv> "map (\<lambda>i. fill_holes (Cs ! i) (sss ! i)) [0..<length Cs]"
    def ts' \<equiv> "map (\<lambda>i. fill_holes (Cs ! i) (tss ! i)) [0..<length Cs]"
    def us' \<equiv> "map (\<lambda>i. fill_holes (Ds ! i) (uss ! i)) [0..<length Ds]"
    def vs' \<equiv> "map (\<lambda>i. fill_holes (Ds ! i) (vss ! i)) [0..<length Ds]"

    have [simp]: "listsum (map num_holes Cs) = length ss"
      using s by (auto simp: C len_Cs num_holes_fill_holes_mctxt dest: eqfE)
    have [simp]: "listsum (map num_holes Ds) = length us"
      using s' by (auto simp: D len_Ds num_holes_fill_holes_mctxt dest: eqfE)

    have len_ss: "length ss = listsum (map num_holes Cs)"
      using s by (auto simp: C len_Cs num_holes_fill_holes_mctxt dest: eqfE)
    have len_us: "length us = listsum (map num_holes Ds)"
      using s' by (auto simp: D len_Ds num_holes_fill_holes_mctxt dest: eqfE)
    have len_vs: "length vs = listsum (map num_holes Ds)"
      using s' by (auto simp: D len_Ds num_holes_fill_holes_mctxt dest: eqfE)

    have ss'_i: "\<And>i. i < length Cs \<Longrightarrow> ss' ! i = fill_holes (Cs ! i) (sss ! i)"
      and us'_i: "\<And>i. i < length Cs \<Longrightarrow> us' ! i = fill_holes (Ds ! i) (uss ! i)"
      and ts'_i: "\<And>i. i < length Cs \<Longrightarrow> ts' ! i = fill_holes (Cs ! i) (tss ! i)"
      and vs'_i: "\<And>i. i < length Cs \<Longrightarrow> vs' ! i = fill_holes (Ds ! i) (vss ! i)"
      using len_Cs len_Ds by (auto simp: ss'_def us'_def ts'_def vs'_def)

    have [simp]: "\<And>i. i < length Cs \<Longrightarrow> length (vss ! i) = length (uss ! i)"
      using len_Cs and len_Ds and len_vs by simp

    have "s = fill_holes E ss'"
      using fill_holes_mctxt_fill_holes [of Cs E ss]
      by (simp add: s_left ss'_def len_Cs C [symmetric])
    moreover
    have "s = fill_holes E us'"
      using fill_holes_mctxt_fill_holes [of Ds E us]
      by (simp add: s_right us'_def len_Ds D [symmetric])
    ultimately
    have us'_ss': "us' = ss'"
      using len_Cs and len_Ds
      by (intro fill_holes_inj [of E]) (simp_all add: ss'_def us'_def)

    have "\<forall>i < length ts'. \<exists>v. (ts' ! i, v) \<in> epar_n R n \<and> (vs' ! i, v) \<in> epar_n R m"
      (is "\<forall>i < length ts'. \<exists>v. ?P i v")
    proof (intro allI impI)
      fix i
      assume [simp]: "i < length ts'"
      then have [simp]: "i < length ss'" "i < length vs'" "i < length us'"
        using len_Cs and len_Ds by (simp_all add: us'_ss' ts'_def ss'_def vs'_def)
      then have [simp]: "i < length Cs" "i < length Ds"
        using len_Cs and len_Ds by (simp_all add: ss'_def)
      then have "i < length (inf_mctxt_args C D)" by (simp add: Cs_def)
      moreover have "(C, D) \<in> comp_mctxt" using s and s' by (metis eqf_comp_mctxt)
      ultimately have "Cs ! i = MHole \<or> Ds ! i = MHole"
        using inf_mctxt_args_MHole by (auto simp: Cs_def Ds_def)
      then show "\<exists>v. ?P i v"
      proof
        assume [simp]: "Cs ! i = MHole"
        def j \<equiv> "listsum (map length (take i tss))"
        have "j < length ss"
          unfolding len_ss
          by (subst id_take_nth_drop [OF \<open>i < length Cs\<close>])
             (simp add: take_map [symmetric] drop_map [symmetric] j_def)

        have "ts' ! i = tss ! i ! 0" by (auto simp: ts'_def intro: fill_holes_MHole)
        then have [simp]: "ts' ! i = ts ! j"
          using partition_by_nth_nth [of i ts "map num_holes Cs" 0] by (simp add: j_def)

        have "ss' ! i = sss ! i ! 0" by (auto simp: ss'_def intro: fill_holes_MHole)
        then have [simp]: "ss' ! i = ss ! j"
          using partition_by_nth_nth [of i ss "map num_holes Cs" 0]
          by (auto simp: j_def
            intro!: arg_cong [of _ _ "nth ss"] arg_cong [of _ _ listsum] arg_cong [of _ _ "take i"] nth_map_conv)

        have "\<forall>i < length (concat uss). (concat uss ! i, concat vss ! i) \<in> epar_n R n"
          using us_epar by simp
        moreover have "i < length uss"
          using len_Cs and len_Ds by simp
        ultimately have epar_uss: "\<forall>j < length (uss ! i). (uss ! i ! j, vss ! i ! j) \<in> epar_n R n"
          using concat_nth_length [of i uss] and concat_nth_length [of i vss]
          by (auto simp: concat_nth [symmetric, OF _ _ refl] take_map [symmetric])
        then have epar_us': "(us' ! i, vs' ! i) \<in> epar_n R n"
          using len_Cs len_Ds by (intro epar_n_mctxt [of _ "Ds ! i"]) (auto simp: us'_i vs'_i)

        have uss: "\<forall>j < length (uss ! i). (uss ! i ! j, vss ! i ! j) \<in> trs_n R (Suc n') \<union> (cstep_n R n')\<^sup>*"
          using concat_nth_length [of i uss] and concat_nth_length [of i vss] and us
          by (auto simp: concat_nth [symmetric, OF _ _ refl] take_map [symmetric])

        have [simp]: "ss ! j = us' ! i" by (simp add: us'_ss')

        have epar_ss': "(ss' ! i, ts' ! i) \<in> epar_n R m"
          using \<open>j < length ss\<close> and ss_epar by auto
        
        have "(ss' ! i, ts' ! i) \<in> trs_n R m \<union> (cstep_n R m')\<^sup>*" using ss and \<open>j < length ss\<close> by auto
        moreover
        { assume "(ss' ! i, ts' ! i) \<in> (cstep_n R m')\<^sup>*"
          then have "(ss' ! i, ts' ! i) \<in> (epar_n R m')\<^sup>*"
            using rtrancl_mono [OF cstep_n_subset_epar_n] by blast
          moreover have "comm ((epar_n R m')\<^sup>*) (epar_n R n)" using 2 by simp
          moreover have "(ss' ! i, vs' ! i) \<in> (epar_n R n)" using epar_us' by simp
          ultimately obtain v where "(ts' ! i, v) \<in> epar_n R n"
            and "(vs' ! i, v) \<in> (epar_n R m')\<^sup>*" by (auto elim: commE)
          then have ?thesis using epar_n_rtrancl_Suc_mono [of R m'] by auto }
        moreover
        { assume trs_m: "(ss' ! i, ts' ! i) \<in> trs_n R m"
          have ?thesis
          proof (cases "Ds ! i = MHole")
            case True [simp]
            def k \<equiv> "listsum (map length (take i uss))"
            have "k < length us"
            unfolding len_us
            by (subst id_take_nth_drop [OF \<open>i < length Ds\<close>])
               (simp add: k_def take_map [symmetric])
  
            have "vs' ! i = vss ! i ! 0" by (auto simp: vs'_def intro: fill_holes_MHole)
            then have [simp]: "vs' ! i = vs ! k"
              using partition_by_nth_nth [of i vs "map num_holes Ds" 0]
              by (simp add: k_def take_map [symmetric])
    
            have "us' ! i = uss ! i ! 0" by (auto simp: us'_def intro: fill_holes_MHole)
            then have [simp]: "us' ! i = us ! k"
              using partition_by_nth_nth [of i us "map num_holes Ds" 0] by (simp add: k_def)

            have "(us' ! i, vs' ! i) \<in> trs_n R n \<union> (cstep_n R n')\<^sup>*"
              using us and \<open>k < length us\<close> by auto
            moreover
            { assume "(us' ! i, vs' ! i) \<in> (cstep_n R n')\<^sup>*"
              then have "(us' ! i, vs' ! i) \<in> (epar_n R n')\<^sup>*"
                using rtrancl_mono [OF cstep_n_subset_epar_n] by blast
              moreover have "comm ((epar_n R n')\<^sup>*) (epar_n R m)" using 2 by simp
              moreover have "(us' ! i, ts' ! i) \<in> (epar_n R m)" using epar_ss' by simp
              ultimately obtain v where "(ts' ! i, v) \<in> (epar_n R n')\<^sup>*"
                and "(vs' ! i, v) \<in> epar_n R m" by (auto elim: commE)
              then have ?thesis using epar_n_rtrancl_Suc_mono [of R n'] by auto }
            moreover
            { assume "(us' ! i, vs' ! i) \<in> trs_n R n"
              with trs_m have ?thesis using 1 2 3 by (auto elim: trs_n_peak) }
            ultimately show ?thesis by blast
          next
            case False
            have *: "us' ! i =\<^sub>f (Ds ! i, uss ! i)"
              using len_Cs and len_Ds by (auto simp: us'_i)
            have len_vss: "length (vss ! i) = length (uss ! i)" by simp
            from epar_n_varpeak [OF * False len_vss trs_m [simplified] uss 2 3]
              show ?thesis by (simp add: vs'_i)
          qed }
        ultimately show ?thesis by blast
      next
        assume [simp]: "Ds ! i = MHole"
        def j \<equiv> "listsum (map length (take i vss))"
        have "j < length us"
          unfolding len_us
          by (subst id_take_nth_drop [OF \<open>i < length Ds\<close>])
             (simp add: take_map [symmetric] drop_map [symmetric] j_def)

        have "vs' ! i = vss ! i ! 0" by (auto simp: vs'_def intro: fill_holes_MHole)
        then have [simp]: "vs' ! i = vs ! j"
          using partition_by_nth_nth [of i vs "map num_holes Ds" 0] by (simp add: j_def)

        have "us' ! i = uss ! i ! 0" by (auto simp: us'_def intro: fill_holes_MHole)
        then have [simp]: "us' ! i = us ! j"
          using partition_by_nth_nth [of i us "map num_holes Ds" 0]
          by (auto simp: j_def
            intro!: arg_cong [of _ _ "nth us"] arg_cong [of _ _ listsum] arg_cong [of _ _ "take i"] nth_map_conv)

        have "\<forall>i < length (concat sss). (concat sss ! i, concat tss ! i) \<in> epar_n R m"
          using ss_epar by simp
        moreover have "i < length sss"
          using len_Cs and len_Ds by simp
        ultimately have epar_uss: "\<forall>j < length (sss ! i). (sss ! i ! j, tss ! i ! j) \<in> epar_n R m"
          using concat_nth_length [of i sss] and concat_nth_length [of i tss]
          by (auto simp: concat_nth [symmetric, OF _ _ refl] take_map [symmetric])
        then have epar_ss': "(ss' ! i, ts' ! i) \<in> epar_n R m"
          using len_Cs len_Ds by (intro epar_n_mctxt [of _ "Cs ! i"]) (auto simp: ss'_i ts'_i)

        have sss: "\<forall>j < length (sss ! i). (sss ! i ! j, tss ! i ! j) \<in> trs_n R (Suc m') \<union> (cstep_n R m')\<^sup>*"
          using concat_nth_length [of i sss] and concat_nth_length [of i tss] and ss
          by (auto simp: concat_nth [symmetric, OF _ _ refl] take_map [symmetric])

        have [simp]: "us ! j = ss' ! i" by (simp add: us'_ss' [symmetric])

        have epar_us': "(us' ! i, vs' ! i) \<in> epar_n R n"
          using \<open>j < length us\<close> and us_epar by auto

        have "(us' ! i, vs' ! i) \<in> trs_n R n \<union> (cstep_n R n')\<^sup>*" using us and \<open>j < length us\<close> by auto
        moreover
        { assume "(us' ! i, vs' ! i) \<in> (cstep_n R n')\<^sup>*"
          then have "(us' ! i, vs' ! i) \<in> (epar_n R n')\<^sup>*"
            using rtrancl_mono [OF cstep_n_subset_epar_n] by blast
          moreover have "comm ((epar_n R n')\<^sup>*) (epar_n R m)" using 2 by simp
          moreover have "(us' ! i, ts' ! i) \<in> (epar_n R m)" using epar_ss' by simp
          ultimately obtain v where "(vs' ! i, v) \<in> epar_n R m"
            and "(ts' ! i, v) \<in> (epar_n R n')\<^sup>*" by (auto elim: commE)
          then have ?thesis using epar_n_rtrancl_Suc_mono [of R n'] by auto }
        moreover
        { assume trs_n: "(us' ! i, vs' ! i) \<in> trs_n R n"
          have ?thesis
          proof (cases "Cs ! i = MHole")
            case True [simp]
            def k \<equiv> "listsum (map length (take i sss))"
            have "k < length ss"
            unfolding len_ss
            by (subst id_take_nth_drop [OF \<open>i < length Cs\<close>])
               (simp add: k_def take_map [symmetric])

            have "ts' ! i = tss ! i ! 0" by (auto simp: ts'_def intro: fill_holes_MHole)
            then have [simp]: "ts' ! i = ts ! k"
              using partition_by_nth_nth [of i ts "map num_holes Cs" 0]
              by (simp add: k_def take_map [symmetric])

            have "ss' ! i = sss ! i ! 0" by (auto simp: ss'_def intro: fill_holes_MHole)
            then have [simp]: "ss' ! i = ss ! k"
              using partition_by_nth_nth [of i ss "map num_holes Cs" 0] by (simp add: k_def)

            have "(ss' ! i, ts' ! i) \<in> trs_n R m \<union> (cstep_n R m')\<^sup>*"
              using ss and \<open>k < length ss\<close> by auto
            moreover
            { assume "(ss' ! i, ts' ! i) \<in> (cstep_n R m')\<^sup>*"
              then have "(ss' ! i, ts' ! i) \<in> (epar_n R m')\<^sup>*"
                using rtrancl_mono [OF cstep_n_subset_epar_n] by blast
              moreover have "comm ((epar_n R m')\<^sup>*) (epar_n R n)" using 2 by simp
              moreover have "(ss' ! i, vs' ! i) \<in> (epar_n R n)" using epar_us' by simp
              ultimately obtain v where "(vs' ! i, v) \<in> (epar_n R m')\<^sup>*"
                and "(ts' ! i, v) \<in> epar_n R n" by (auto elim: commE)
              then have ?thesis using epar_n_rtrancl_Suc_mono [of R m'] by auto }
            moreover
            { assume "(ss' ! i, ts' ! i) \<in> trs_n R m"
              with trs_n have ?thesis using 1 2 3 by (auto elim: trs_n_peak) }
            ultimately show ?thesis by blast
          next
            case False
            have *: "ss' ! i =\<^sub>f (Cs ! i, sss ! i)"
              using len_Cs and len_Ds by (auto simp: ss'_i)
            have len_tss: "length (tss ! i) = length (sss ! i)" by simp
            from epar_n_varpeak [OF * False len_tss trs_n [simplified] sss 2 3]
              show ?thesis by (auto simp: ts'_i)
          qed }
        ultimately show ?thesis by blast
      qed
    qed
    then obtain v where "\<forall>i < length ts'. ?P i (v i)" unfolding choice_iff' by blast
    moreover def ws \<equiv> "map v [0 ..< length ts']"
    ultimately have *: "\<forall>i < length ts'. ?P i (ws ! i)" by (simp)

    have [simp]: "length ts' = length vs'"
      by (simp add: vs'_def ts'_def len_Cs [symmetric] len_Ds)

    def w \<equiv> "fill_holes E ws"

    have [simp]: "num_holes (fill_holes_mctxt E Cs) = length ss"
      using s by (auto simp: C [symmetric] dest!: eqfE)
    have [simp]: "num_holes (fill_holes_mctxt E Ds) = length vs"
      using s' by (auto simp: D [symmetric] dest!: eqfE)
    have "(t, w) \<in> epar_n R n"
      using *
      by (intro epar_n_mctxt [of _ E ts' _ ws])
        (auto simp: ts'_def eq_fill.simps len_Cs fill_holes_mctxt_fill_holes w_def ws_def)
    moreover have "(u, w) \<in> epar_n R m"
      using *
      by (intro epar_n_mctxt [of _ E vs' _ ws])
        (auto simp: vs'_def eq_fill.simps len_Ds fill_holes_mctxt_fill_holes w_def ws_def len_Cs)
    ultimately have ?case by blast }
  ultimately show ?case
    by (cases m n rule: nat.exhaust [case_product nat.exhaust]) simp_all
qed

text \<open>
  Almost orthogonal (modulo infeasibility), properly oriented, right-stable 3-CTRSs
  are level-confluent.
\<close>
lemma level_confluence:
  shows "level_confluent R"
unfolding level_confluent_def
proof (intro allI diamond_imp_CR')
  fix n
  show "cstep_n R n \<subseteq> epar_n R n" by (rule cstep_n_subset_epar_n)
  show "epar_n R n \<subseteq> (cstep_n R n)\<^sup>*" by (rule epar_n_subset_csteps_n)
  show "\<diamond> (epar_n R n)" using comm_epar_n by fast
qed

end

lemma level_confluent_imp_CR:
  assumes "level_confluent R"
  shows "CR (cstep R)"
proof
  fix a b c
  assume "(a, b) \<in> (cstep R)\<^sup>*"
    and "(a, c) \<in> (cstep R)\<^sup>*"
  then obtain n and m where "(a, b) \<in> (cstep_n R n)\<^sup>*"
    and "(a, c) \<in> (cstep_n R m)\<^sup>*" using csteps_imp_csteps_n[of _ _ R] by blast
  then have "(a, b) \<in> (cstep_n R (n + m))\<^sup>*"
    and "(a, c) \<in> (cstep_n R (n + m))\<^sup>*"
    using cstep_n_mono [THEN rtrancl_mono, of _ "n + m" R]
    and le_add1[of n m] and le_add2[of m n] by blast+
  then obtain d where bd: "(b, d) \<in> (cstep_n R (n + m))\<^sup>*"
    and cd: "(c, d) \<in> (cstep_n R (n + m))\<^sup>*" using assms
    by (auto simp: level_confluent_def CR_defs) (metis joinE)
  then have "(b, d) \<in> (cstep R)\<^sup>*" and "(c, d) \<in> (cstep R)\<^sup>*"
  by (metis (erased, hide_lams) contra_subsetD cstep_iff rtrancl_mono subrelI)+
  then show "(b, c) \<in> (cstep R)\<^sup>\<down>" by auto
qed

end
