(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
(* 
  Calling external termination tools
*)

signature TERMINATION_TOOL =
sig

  type tool = Path.T (*tmp_in*) -> Path.T (*tmp_out*) -> string (*command*)

  datatype result =
    Success of string (*certificate*)
  | Failure of string (*message*)

  val invoke : tool -> string (*problem*) -> result
  val aprove : tool
  val ttt2 : tool

end

structure Termination_Tool: TERMINATION_TOOL =
struct

type tool = Path.T -> Path.T -> string;

datatype result =
  Success of string
| Failure of string;

fun invoke tool problem =
  let
    val {output, redirected_output, return_code} =
      Cache_IO.run tool problem;
  in
    if return_code = 0
    then Success (cat_lines redirected_output)
    else Failure (cat_lines output)
  end;

fun aprove in_path out_path =
  let
    val aprove_cmd = getenv "APROVE_CMD";
    val _ = aprove_cmd <> "" orelse error "Aprove not configured.";
  in
    (File.shell_quote aprove_cmd :: map File.shell_path [in_path, out_path] @ ["2>&1"])
    |> space_implode " "
  end;

fun ttt2 in_path out_path =
  let
    val ttt2_cmd = getenv "TTT2_CMD";
    val _ = ttt2_cmd <> "" orelse error "ttt2 not configured.";
  in
    File.shell_quote ttt2_cmd
      :: (File.shell_path in_path)
      :: ">"
      :: (File.shell_path out_path)
    |> space_implode " "
  end;


end
